# Copyright (C) 2020,2025 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import base64
import functools
import logging
from collections.abc import Collection
from contextlib import suppress
from typing import TYPE_CHECKING, Any, Literal, cast
from urllib.parse import quote, urlparse

import emoji
from cachetools import TTLCache, keys
from jwcrypto import jwk, jws
from OpenSSL.crypto import (
    FILETYPE_ASN1,
    FILETYPE_PEM,
    X509,
    X509Store,
    X509StoreContext,
    dump_certificate,
    load_certificate,
)
from synapse.api.constants import (
    AccountDataTypes,
    Direction,
    EventContentFields,
    EventTypes,
    HistoryVisibility,
    JoinRules,
    Membership,
    RoomCreationPreset,
)
from synapse.api.errors import SynapseError
from synapse.api.filtering import Filter
from synapse.api.room_versions import RoomVersion
from synapse.config import ConfigError
from synapse.config._base import Config
from synapse.events import EventBase
from synapse.handlers.pagination import SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME
from synapse.http.client import BaseHttpClient
from synapse.http.proxyagent import ProxyAgent
from synapse.http.server import JsonResource
from synapse.logging.context import make_deferred_yieldable
from synapse.module_api import NOT_SPAM, ModuleApi, errors
from synapse.server import HomeServer
from synapse.storage.database import LoggingTransaction
from synapse.types import (
    Requester,
    ScheduledTask,
    StateMap,
    TaskStatus,
    UserID,
    create_requester,
)
from synapse.types.handlers import ShutdownRoomParams
from synapse.types.state import StateFilter
from synapse.util.metrics import measure_func
from twisted.internet.defer import Deferred
from twisted.internet.ssl import PrivateCertificate, optionsForClientTLS, platformTrust
from twisted.web.client import HTTPConnectionPool
from twisted.web.iweb import IAgent, IPolicyForHTTPS
from zope.interface import implementer

from synapse_invite_checker.config import DefaultPermissionConfig, InviteCheckerConfig
from synapse_invite_checker.permissions import InviteCheckerPermissionsHandler
from synapse_invite_checker.rest.messenger_info import (
    INFO_API_PREFIX,
    MessengerFindByIkResource,
    MessengerInfoResource,
    MessengerIsInsuranceResource,
)
from synapse_invite_checker.types import (
    EpaRoomTimestampResults,
    FederationList,
    TimType,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from synapse.storage.roommember import RoomsForUser

# We need to access the private API in some places, in particular the store and the homeserver
# ruff: noqa: SLF001


def cached(cache):
    """Simplified cached decorator from cachetools, that allows calling an async function."""

    def decorator(func):
        async def wrapper(*args, **kwargs):
            k = keys.hashkey(*args, **kwargs)
            with suppress(KeyError):
                return cache[k]

            v = await func(*args, **kwargs)

            with suppress(ValueError):
                cache[k] = v

            return v

        def cache_clear():
            cache.clear()

        wrapper.cache = cache
        wrapper.cache_clear = cache_clear

        return functools.update_wrapper(wrapper, func)

    return decorator


@implementer(IPolicyForHTTPS)
class MtlsPolicy:
    def __init__(self, config: InviteCheckerConfig):
        super().__init__()

        self.url = urlparse(config.federation_list_url)

        # Handle different cases based on URL scheme and mTLS requirements
        client_cert = None
        if config.federation_list_client_cert:
            # If a certificate is provided, always use it
            with open(config.federation_list_client_cert) as file:
                content = file.read()

            client_cert = PrivateCertificate.loadPEM(content)
        elif self.url.scheme == "https" and config.federation_list_require_mtls:
            # HTTPS with required mTLS but no certificate
            msg = "No mtls cert and scheme is https with mTLS required"
            raise Exception(msg)
        elif self.url.scheme not in ("http", "https"):
            # Neither HTTP nor HTTPS
            msg = "URL scheme must be either http or https"
            raise Exception(msg)

        self.options = optionsForClientTLS(
            self.url.hostname, platformTrust(), clientCertificate=client_cert
        )

    def creatorForNetloc(self, hostname: bytes, port: int):
        if (
            self.url.hostname
            and self.url.hostname.encode("utf-8") != hostname
            or self.url.port != port
        ):
            logger.error(
                "Destination mismatch: %r:%r != %r:%r",
                self.url.hostname,
                self.url.port,
                hostname,
                port,
            )
            msg = "Invalid connection attempt by MTLS Policy"
            raise Exception(msg)
        return self.options


class FederationAllowListClient(BaseHttpClient):
    """Custom http client since we need to pass a custom agent to enable mtls"""

    def __init__(
        self,
        hs: HomeServer,
        config: InviteCheckerConfig,
        # We currently assume the configured endpoint is always trustworthy and bypass the proxy
        # ip_allowlist: Optional[IPSet] = None,
        # ip_blocklist: Optional[IPSet] = None,
        # use_proxy: bool = False,
    ):
        super().__init__(hs)

        pool = HTTPConnectionPool(self.reactor)

        mstls_policy = MtlsPolicy(config)
        proxy_agent = ProxyAgent(
            reactor=self.reactor,
            proxy_reactor=hs.get_reactor(),
            connectTimeout=15,
            contextFactory=cast(IPolicyForHTTPS, mstls_policy),
            pool=pool,
        )
        self.agent = cast(IAgent, proxy_agent)


BASE_API_PREFIX = "/_synapse/client/com.famedly/tim"


class InviteChecker:
    __version__ = "0.4.12"

    def __init__(self, config: InviteCheckerConfig, api: ModuleApi):
        self.api = api
        # This needs to be on the Class itself so that metrics functions that measure
        # requests and database calls will function. Specifically for @measure_func
        self.server_name = api.server_name
        self.clock = api._hs.get_clock()

        self.config = config
        # Can not do this as part of parse_config() as there is no access to the server
        # name yet
        self.config.default_permissions.maybe_update_server_exceptions(
            self.api._hs.config.server.server_name
        )

        self.federation_list_client = FederationAllowListClient(api._hs, self.config)

        self.api.register_spam_checker_callbacks(user_may_invite=self.user_may_invite)
        self.api.register_spam_checker_callbacks(
            user_may_join_room=self.user_may_join_room
        )
        self.api.register_third_party_rules_callbacks(
            on_create_room=self.on_create_room
        )
        self.api.register_third_party_rules_callbacks(
            on_upgrade_room=self.on_upgrade_room
        )
        self.api.register_third_party_rules_callbacks(
            check_event_allowed=self.check_event_allowed
        )
        self.api.register_spam_checker_callbacks(
            check_login_for_spam=self.check_login_for_spam
        )

        # Make sure this doesn't get initialized until after the default permissions
        # were potentially modified to account for the local server template
        self.permissions_handler = InviteCheckerPermissionsHandler(
            self.api,
            self.config,
            self.is_domain_insurance,
            self.config.default_permissions,
        )

        self.task_scheduler = api._hs.get_task_scheduler()

        if (
            self.config.room_scan_run_interval_ms > 0
            and self.api.should_run_background_tasks()
        ):
            # The docstring for 'looping_background_call()' is slightly incorrect
            # > Waits msec initially before calling f for the first time.
            # Should be
            # > Calls f after waiting msec, then repeats. This is an inexact, "best effort"
            # > figure when the reactor/event loop is under heavy load
            self.api.looping_background_call(
                self.room_scan, self.config.room_scan_run_interval_ms
            )

        if self.config.tim_type == TimType.PRO:
            # The TiMessengerInformation API resource
            self.resource = JsonResource(api._hs)
            MessengerInfoResource(self.api, self.config).register(self.resource)
            MessengerIsInsuranceResource(
                self.api, self.config, self.is_domain_insurance
            ).register(self.resource)
            MessengerFindByIkResource(
                self.api, self.config, self._fetch_federation_list
            ).register(self.resource)
            self.api.register_web_resource(INFO_API_PREFIX, self.resource)

        self.api._clock.call_when_running(self.after_startup)

        logger.info("Module initialized at %s", BASE_API_PREFIX)

    @staticmethod
    def parse_config(config: dict[str, Any]) -> InviteCheckerConfig:
        logger.error("PARSE CONFIG")

        _default_permissions = config.get("default_permissions", {})
        # The default for permissions when not declared in the configuration as a
        # template is now part of the PermissionConfig class. If you need to change a
        # default, do so there.
        _config = InviteCheckerConfig(DefaultPermissionConfig(**_default_permissions))

        _config.title = config.get("title", _config.title)
        _config.description = config.get("description", _config.description)
        _config.contact = config.get("contact", _config.contact)
        _config.federation_list_client_cert = config.get(
            "federation_list_client_cert", ""
        )
        _config.federation_list_url = config.get("federation_list_url", "")
        _config.federation_list_require_mtls = config.get(
            "federation_list_require_mtls", _config.federation_list_require_mtls
        )
        _config.gematik_ca_baseurl = config.get("gematik_ca_baseurl", "")

        if not _config.federation_list_url or not _config.gematik_ca_baseurl:
            msg = "Incomplete federation list config"
            raise Exception(msg)

        if (
            _config.federation_list_url.startswith("https")
            and _config.federation_list_require_mtls
            and not _config.federation_list_client_cert
        ):
            msg = "Federation list config requires an mtls (PEM) cert for https connections when mTLS is required"
            raise Exception(msg)

        # Validate federation_list_require_mtls is a boolean
        if not isinstance(_config.federation_list_require_mtls, bool):
            msg = "`federation_list_require_mtls` must be a boolean"
            raise ConfigError(msg)

        # Check that the configuration is defined. This allows a grace period for
        # migration. For now, just issue a warning in the logs. The default of 'pro'
        # is set inside InviteCheckerConfig
        _tim_type = config.get("tim-type", "").lower()
        if not _tim_type:
            logger.warning(
                "Please remember to set `tim-type` in your configuration. Defaulting to 'Pro' mode"
            )

        elif _tim_type == "epa":
            _config.tim_type = TimType.EPA
        elif _tim_type == "pro":
            _config.tim_type = TimType.PRO
        else:
            msg = "`tim-type` setting is not a recognized value. Please fix."
            raise ConfigError(msg)

        _allowed_room_versions = config.get("allowed_room_versions", ["9", "10"])
        if not _allowed_room_versions or not isinstance(_allowed_room_versions, list):
            msg = "Allowed room versions must be formatted as a list."
            raise ConfigError(msg)

        _config.allowed_room_versions = [
            # Coercing into a string, in case the yaml loader thought it was an int
            str(_room_ver)
            for _room_ver in _allowed_room_versions
        ]

        run_interval = Config.parse_duration(config.get("room_scan_run_interval", "1h"))
        clamp_minimum_to = Config.parse_duration("1m")

        # If 'room_scan_run_interval' is not set to 0 for disabling the room scan
        # completely, make sure anything less than the minimum of 1 minute is ignored
        _config.room_scan_run_interval_ms = (
            max(run_interval, clamp_minimum_to) if run_interval > 0 else run_interval
        )

        insured_room_scan_section = config.get("insured_only_room_scan", {})
        if not isinstance(insured_room_scan_section, dict):
            msg = "`insured_only_room_scan` should be configured as a dictionary"
            raise ConfigError(msg)

        # Only default enable this room scan if in EPA mode
        enable_insured_room_scan = insured_room_scan_section.get(
            "enabled", _config.tim_type == TimType.EPA
        )
        # But also prevent it running in PRO mode completely
        enable_insured_room_scan = (
            False if _config.tim_type == TimType.PRO else enable_insured_room_scan
        )
        _config.insured_room_scan_options.enabled = enable_insured_room_scan

        epa_room_grace_period = Config.parse_duration(
            insured_room_scan_section.get("grace_period", "1w")
        )
        # For now, Gematik spec requires invites to count as room participation, but this
        # can lead to a room never being considered for epa user kicking. Allow an
        # enforceable maximum on this that is optional
        epa_room_invites_grace_period = Config.parse_duration(
            insured_room_scan_section.get("invites_grace_period", 0)
        )

        _config.insured_room_scan_options.grace_period_ms = epa_room_grace_period
        _config.insured_room_scan_options.invites_grace_period_ms = (
            epa_room_invites_grace_period
        )

        # This option is considered for all server modes unlike 'insured_only_room_scan'
        inactive_room_scan_section = config.get("inactive_room_scan", {})
        if not isinstance(inactive_room_scan_section, dict):
            msg = "`inactive_room_scan` should be formatted as a dictionary"
            raise ConfigError(msg)

        enable_inactive_room_scan = inactive_room_scan_section.get("enabled", True)
        _config.inactive_room_scan_options.enabled = enable_inactive_room_scan

        # "26w" calculates as 6 months
        inactive_room_scan_grace_period = Config.parse_duration(
            inactive_room_scan_section.get("grace_period", "26w")
        )
        _config.inactive_room_scan_options.grace_period_ms = (
            inactive_room_scan_grace_period
        )

        _override_public_room_federation = config.get(
            "override_public_room_federation", _config.override_public_room_federation
        )
        if not isinstance(_override_public_room_federation, bool):
            msg = "`override_public_room_federation` must be a boolean"
            raise ConfigError(msg)
        _config.override_public_room_federation = _override_public_room_federation

        _prohibit_world_readable_rooms = config.get(
            "prohibit_world_readable_rooms", _config.prohibit_world_readable_rooms
        )
        if not isinstance(_prohibit_world_readable_rooms, bool):
            msg = "`prohibit_world_readable_rooms` must be a boolean"
            raise ConfigError(msg)
        _config.prohibit_world_readable_rooms = _prohibit_world_readable_rooms

        _block_invites_into_dms = config.get(
            "block_invites_into_dms", _config.block_invites_into_dms
        )
        _config.block_invites_into_dms = _block_invites_into_dms

        _limit_reactions = config.get("limit_reactions", _config.limit_reactions)
        if not isinstance(_limit_reactions, bool):
            msg = "`limit_reactions` must be a boolean"
            raise ConfigError(msg)
        _config.limit_reactions = _limit_reactions

        return _config

    def after_startup(self) -> None:
        d = Deferred.fromCoroutine(self._after_startup())
        # Whenever a raw Deferred is created and ran, it needs to be wrapped into a log
        # context. 'make_deferred_yieldable()' does that for us
        make_deferred_yieldable(d)

    async def _after_startup(self) -> None:
        """
        To be called when the reactor is running. Validates that the epa setting matches
        the insurance setting in the federation list.
        """
        try:
            fed_list = await self._fetch_federation_list()
            if self.config.tim_type == TimType.EPA and not fed_list.is_insurance(
                self.api._hs.config.server.server_name
            ):
                logger.warning(
                    "This server has enabled ePA Mode in its config, but is not found on "
                    "the Federation List as an Insurance Domain!"
                )
        except Exception as e:  # noqa: BLE001
            logger.warning(
                "The server had an issue retrieving the Federation List: %r", e
            )

    async def _raw_federation_list_fetch(self) -> str:
        resp = await self.federation_list_client.get_raw(
            self.config.federation_list_url
        )
        return resp.decode()

    async def _raw_gematik_root_ca_fetch(self) -> dict:
        return await self.api._hs.get_proxied_http_client().get_json(
            f"{self.config.gematik_ca_baseurl}/ECC/ROOT-CA/roots.json"
        )

    async def _raw_gematik_intermediate_cert_fetch(self, cn: str) -> bytes:
        return await self.api._hs.get_proxied_http_client().get_raw(
            f"{self.config.gematik_ca_baseurl}/ECC/SUB-CA/{quote(cn.replace(' ', '_'), safe='')}.der"
        )

    def _load_cert_b64(self, cert: str) -> X509:
        return load_certificate(FILETYPE_ASN1, base64.b64decode(cert))

    @cached(cache=TTLCache(maxsize=1, ttl=60 * 60))
    async def _fetch_federation_list(
        self,
    ) -> FederationList:
        """
        Fetch the raw data for the federation list, verify it is authentic and parse
        the data into a usable format

        Returns:
            a FederationList object

        """
        raw_list = await self._raw_federation_list_fetch()
        jws_verify = jws.JWS()
        jws_verify.deserialize(raw_list, alg="BP256R1")
        jws_verify.allowed_algs = ["BP256R1"]

        jwskey = self._load_cert_b64(jws_verify.jose_header["x5c"][0])

        # TODO(Nico): Fetch the ca only once a week
        store = X509Store()
        roots = await self._raw_gematik_root_ca_fetch()
        for r in roots:
            rawcert = r["cert"]
            if rawcert:
                store.add_cert(self._load_cert_b64(rawcert))

        chain = load_certificate(
            FILETYPE_ASN1,
            await self._raw_gematik_intermediate_cert_fetch(
                jwskey.get_issuer().CN or ""
            ),
        )
        store_ctx = X509StoreContext(store, jwskey, chain=[chain])
        store_ctx.verify_certificate()

        key = jwk.JWK.from_pem(dump_certificate(FILETYPE_PEM, jwskey))

        jws_verify.verify(key, alg="BP256R1")

        if jws_verify.payload is None:
            msg = "Empty federation list"
            raise Exception(msg)

        # Validate incoming, potentially incomplete or corrupt data
        return FederationList.model_validate_json(jws_verify.payload)

    async def is_domain_allowed(self, domain: str) -> bool:
        """
        See if a domain is found on the Federation List. If it was not, refetch the list
        to try again.
        """
        fed_list = await self._fetch_federation_list()
        if fed_list.allowed(domain):
            return True

        # Per A_25537:
        # The domain wasn't found but the list may have changed since the last look.
        # Re-fetch the list and try again. See:
        # https://gemspec.gematik.de/docs/gemSpec/gemSpec_TI-M_Basis/gemSpec_TI-M_Basis_V1.1.1/#A_25537
        self._fetch_federation_list.cache_clear()
        fed_list = await self._fetch_federation_list()
        return fed_list.allowed(domain)

    async def is_domain_insurance(self, domain: str) -> bool:
        """See if a domain was considered an insurance domain per the Federation List"""
        fed_list = await self._fetch_federation_list()
        return fed_list.is_insurance(domain)

    async def on_upgrade_room(
        self, _: Requester, room_version: RoomVersion, is_requester_admin: bool = False
    ) -> None:
        if (
            not is_requester_admin
            and room_version.identifier not in self.config.allowed_room_versions
        ):
            raise SynapseError(
                400,
                f"Room version ('{room_version}') not allowed",
                errors.Codes.FORBIDDEN,
            )

    async def user_may_join_room(
        self, user: str, room_id: str, is_invited: bool
    ) -> Literal["NOT_SPAM"] | errors.Codes:
        """
        This is used to check that a local user can join a room. Invites to remote
        public rooms MUST be denied. Invites to local rooms are allowed(unless it is an
        EPA server, in which case it should not get here)
        Args:
            user:
            room_id:
            is_invited:

        Returns:

        """
        if not is_invited:
            # Do we have the creation event of the room state?
            state_mapping: StateMap[EventBase] = (
                await self.api._storage_controllers.state.get_current_state(
                    room_id,
                    StateFilter.from_types(
                        [(EventTypes.Create, None), (EventTypes.JoinRules, None)]
                    ),
                )
            )

            creation_event = state_mapping.get((EventTypes.Create, ""))
            if not creation_event:
                # This happens because we do not have the state of the room. If this was
                # an invite(which includes 'invite_room_state') we would not be here. It
                # is highly likely that this means the room is remote. Since remote
                # rooms with no invite are not allowed, deny the request. Local rooms
                # that do not have state should be in the act of purging, in which case
                # we do not want to allow that join anyway.
                logger.debug(
                    "Denying join of '%s' to room '%s' because local server has no state(which represents a remote room)",
                    user,
                    room_id,
                )
                return errors.Codes.FORBIDDEN

            # There was no invite, but we already have the state of the room. Deny
            # public rooms if the room's creator's domain isn't the same as the local
            # server
            room_creator = UserID.from_string(creation_event.sender)
            if room_creator.domain != self.server_name:
                join_rules = state_mapping.get((EventTypes.JoinRules, ""))
                # all rooms should have join_rules, make sure
                if join_rules is None:
                    logger.warning(
                        "Room state of '%s' does not contain 'join_rules", room_id
                    )
                    return errors.Codes.FORBIDDEN

                if join_rules.content["join_rule"] == JoinRules.PUBLIC:
                    # There are no public remote rooms.
                    logger.debug(
                        "Forbidding join of '%s' to remote PUBLIC room '%s'",
                        user,
                        room_id,
                    )
                    return errors.Codes.FORBIDDEN

            # Room was created by a local user
            return NOT_SPAM

        else:
            # Try and see if the invite event had any initial room state data. For now,
            # this requires a database call, but if https://github.com/element-hq/synapse/issues/18230
            # becomes a thing, we won't need it anymore. It is possible that room_data
            # can be None. Logically however, it would only be None if there was no
            # invite. Since those conditions are checked for above, this should be safe
            room_data = await self.api._store.get_invite_for_local_user_in_room(
                user, room_id
            )
            if room_data is None:
                # If for some reason this data is missing, deny and bail. Someone is doing
                # something fishy
                logger.warning(
                    "Forbidding join of '%s' to room '%s' because invite data could not be found",
                    user,
                    room_id,
                )
                return errors.Codes.FORBIDDEN

            invite_event = await self.api._store.get_event(room_data.event_id)

            create_event_senders_domain = None
            is_public = True

            # Sort out the conditions
            for _event in invite_event.unsigned.get("invite_room_state", []):
                if (
                    _event["type"] == EventTypes.JoinRules
                    and _event["content"]["join_rule"] != JoinRules.PUBLIC
                ):
                    is_public = False
                if _event["type"] == EventTypes.Create:
                    create_event_senders_domain = UserID.from_string(
                        _event["sender"]
                    ).domain

            if is_public and create_event_senders_domain != self.server_name:
                logger.debug(
                    "Forbidding joining '%s' to invited room '%s' because room is PUBLIC",
                    user,
                    room_id,
                )
                return errors.Codes.FORBIDDEN

            return NOT_SPAM

    async def check_event_allowed(
        self, event: EventBase, context: StateMap[EventBase]
    ) -> tuple[bool, dict | None]:
        """
        Check the 'event' to see if it is allowed to exist. This takes place before
        the event is actually stored.
        Args:
            event: The unpersisted EventBase for the new Event
            context: The StateMapping for the new Event, from just before the new Event

        Returns: tuple of bool, for if the Event is 'allowed' and an optional dict of the
        replacement to use for the new event, in case modification is needed. WARNING:
        this has hazardous potential to break federation, and it is extremely unlikely
        we will ever use it

        Raises: SynapseError(400, M_BAD_JSON) if a reaction annotation has more than a
            single grapheme cluster when this restriction is enabled in settings
        """
        # This call check has many places it can be used, short-circuit out as swiftly
        # as is feasible
        # Never touch anything from another server
        if not self.api.is_mine(event.sender):
            return True, None

        if not event.is_state():
            # Only judge m.reaction when it is not a state event
            if self.config.limit_reactions and event.type == EventTypes.Reaction:
                key: str = event.content["m.relates_to"]["key"]

                # Using `is_emoji()` will not check for emoji variations(such as skin
                # tone or gender identifiers), but `purely_emoji()` does. So, first
                # check that it is an emoji(which will exclude normal numbers and
                # letters) and then check that there is only a single emoji(while
                # accommodating the aforementioned emoji variations). `emoji_list()`
                # does seem to handle the variations correctly(based on some limited
                # testing)
                if not emoji.purely_emoji(key) or len(emoji.emoji_list(key)) > 1:
                    # Normally with this API, it is expected to return a bool to
                    # indicate a failure to comply, however this raises a 403 error with
                    # the FORBIDDEN code, and gematik wants it to be a 400 with BAD_JSON.
                    raise SynapseError(
                        400,
                        "Only single emoji reactions are allowed",
                        errors.Codes.BAD_JSON,
                    )

            # Otherwise, we only check state events
            return True, None
        # Important Note: This callback also runs during room creation, and may end up
        # being appropriate for checking the same things we check in `on_create_room()`.
        # However, this callback is run for *every single event created* and should be
        # kept as "light weight" as possible

        # Forbid "m.room.join_rules" being anything but "private" for EPA, and only being
        # "public" on PRO if "m.federate" is set to True in the creation event
        if (
            event.type == EventTypes.JoinRules
            and event.content["join_rule"] == JoinRules.PUBLIC
        ):
            # EPA gets no public rooms, full stop
            if self.config.tim_type == TimType.EPA:
                return False, None
            creation_event = context.get((EventTypes.Create, ""))
            # `m.federate` defaults to True if unspecified
            if creation_event and creation_event["content"]:
                federated_flag = creation_event["content"].get(
                    EventContentFields.FEDERATE, True
                )
            # Remember to account for the override disabler
            # TODO: fix this possible reference before assignment
            if federated_flag and self.config.override_public_room_federation:
                return False, None

        # If configured, forbid "m.room.history_visibility" to be set as "world_readable"
        elif (
            self.config.prohibit_world_readable_rooms
            and event.type == EventTypes.RoomHistoryVisibility
            and event.content["history_visibility"] == HistoryVisibility.WORLD_READABLE
        ):
            return False, None

        return True, None

    async def on_create_room(
        self,
        requester: Requester,
        request_content: dict[str, Any],
        is_request_admin: bool,
    ) -> None:
        """
        Raise a SynapseError if creating a room should be denied. Currently, this checks
        * invites
        * room version
        * room public-ness via room creation presets
        """
        # visibility can be either "public" or "private". If not included, it defaults to "private"
        room_visibility: str = request_content.get("visibility", "private")
        # preset can be any of "private_chat", "trusted_private_chat" or "public_chat"
        # Do not allow "public_chat". Default is based on setting of visibility
        room_preset: str = request_content.get(
            "preset",
            (
                RoomCreationPreset.PUBLIC_CHAT
                if room_visibility == "public"
                else RoomCreationPreset.PRIVATE_CHAT
            ),
        )
        # Determine based on above that the room is probably public
        is_public = (
            room_preset == RoomCreationPreset.PUBLIC_CHAT or room_visibility == "public"
        )

        if self.config.tim_type == TimType.PRO:
            creation_content = request_content.get("creation_content", {})
            # m.federate defaults to True if unspecified
            can_federate = creation_content.get("m.federate", True)

            if can_federate and is_public:
                if self.config.override_public_room_federation:
                    logger.debug("Overriding `m.room.create` to disable federation")
                    request_content.setdefault("creation_content", {}).update(
                        {"m.federate": False}
                    )
                else:
                    logger.warning(
                        "Room creation with a public room allowed to federate detected."
                    )

        # Forbid EPA servers from creating any kind of public room
        if self.config.tim_type == TimType.EPA and is_public:
            raise SynapseError(
                400,
                "Creation of a public room is not allowed",
                errors.Codes.FORBIDDEN,
            )

        if is_request_admin:
            return

        # Unlike `user_may_invite()`, `on_create_room()` only runs with the inviter being
        # a local user; the invitee can be local/remote. Unfortunately, the spam check module
        # function `user_may_create_room()` only accepts the user creating the room and
        # has no other information provided.

        invite_list: list[str] = request_content.get("invite", [])
        # Per A_25538, only a single additional user may be invited to a room during
        # creation. See:
        # https://gemspec.gematik.de/docs/gemSpec/gemSpec_TI-M_Basis/gemSpec_TI-M_Basis_V1.1.1/#A_25538
        # Interesting potential error here, they display an http error code of 400, but
        # then say to use "M_FORBIDDEN". Pretty sure that is a typo
        if len(invite_list) > 1:
            raise SynapseError(
                403,
                "When creating a room, a maximum of one participant can be invited directly",
                errors.Codes.FORBIDDEN,
            )

        inviter = requester.user.to_string()
        for invitee in invite_list:
            res = await self.user_may_invite(inviter, invitee)
            if res != "NOT_SPAM":
                raise SynapseError(
                    403,
                    f"Room not created as user ({invitee}) is not allowed to be invited",
                    errors.Codes.FORBIDDEN,
                )

        # The room version should always be a string to accommodate arbitrary unstable
        # room versions. If it was not explicitly requested, the homeserver defaults
        # will be used. Make sure to check that instance as well
        room_version: str = request_content.get(
            "room_version", self.api._hs.config.server.default_room_version.identifier
        )

        if room_version not in self.config.allowed_room_versions:
            raise SynapseError(
                400,
                f"Room version ('{room_version}') not allowed",
                errors.Codes.FORBIDDEN,
            )

    async def check_login_for_spam(
        self,
        user_id: str,
        _device_id: str | None,
        _initial_display_name: str | None,
        _request_info: Collection[tuple[str | None, str]],
        _auth_provider_id: str | None = None,
    ) -> Literal["NOT_SPAM"] | errors.Codes:
        # Default permissions are populated automatically when fetching them. This
        # ensures users can see the default permissions in their client when they sign
        # in.
        await self.permissions_handler.get_permissions(user_id)
        return NOT_SPAM

    async def user_may_invite(
        self, inviter: str, invitee: str, room_id: str | None = None
    ) -> Literal["NOT_SPAM"] | errors.Codes:
        # Verify that local users can't invite into their DMs as verified by a few
        # tests in the Testsuite. In the context of calling this directly from
        # `on_create_room()` above, there may not be a room_id yet.
        if self.api.is_mine(inviter) and room_id and self.config.block_invites_into_dms:
            direct = await self.api.account_data_manager.get_global(
                inviter, AccountDataTypes.DIRECT
            )
            if direct:
                for user, roomids in direct.items():
                    if room_id in roomids and user != invitee:
                        # Can't invite to DM!
                        logger.debug(
                            "Preventing invite since %s already has a DM with %s",
                            inviter,
                            invitee,
                        )
                        return errors.Codes.FORBIDDEN

        inviter_domain = UserID.from_string(inviter).domain
        invitee_domain = UserID.from_string(invitee).domain

        # Step 1a, check federation allow list. See:
        # https://gemspec.gematik.de/docs/gemSpec/gemSpec_TI-M_Basis/gemSpec_TI-M_Basis_V1.1.1/#A_25534
        if not (
            await self.is_domain_allowed(inviter_domain)
            and await self.is_domain_allowed(invitee_domain)
        ):
            logger.warning(
                "Discarding invite between domains: (%s) and (%s)",
                inviter_domain,
                invitee_domain,
            )
            return errors.Codes.FORBIDDEN

        # Step 1b
        # Per AF_10233: Deny incoming remote invites if in ePA mode(which means the
        # local user is an 'insured') and if the remote domain is type 'insurance'.
        if await self.is_domain_insurance(
            inviter_domain
        ) and await self.is_domain_insurance(invitee_domain):
            logger.warning(
                "Discarding invite from between insured users: %s and %s",
                inviter,
                invitee,
            )
            return errors.Codes.FORBIDDEN

        # Find out if this is a public room
        # The domains are different, or the first section would have caught it. The same
        # context as above applies, there may not yet be a room_id if this is a room
        # creation in progress
        if room_id and (not self.api.is_mine(inviter) or not self.api.is_mine(invitee)):
            state_mapping: StateMap[EventBase] = (
                await self.api._storage_controllers.state.get_current_state(
                    room_id,
                    StateFilter.from_types([(EventTypes.JoinRules, None)]),
                )
            )
            event = state_mapping.get((EventTypes.JoinRules, ""))
            if event and event.content["join_rule"] == JoinRules.PUBLIC:
                logger.debug(
                    "Forbidding invite to a local public room to a remote user (%s -> %s)",
                    inviter,
                    invitee,
                )
                return errors.Codes.FORBIDDEN

        # Step 2, check invite settings
        # Skip remote users as we can't check their account data
        if self.api.is_mine(invitee):
            if not await self.api.check_user_exists(invitee):
                logger.warning(
                    "Blocking invite to non-existent local user '%s'", invitee
                )
                return errors.Codes.FORBIDDEN

            if not await self.permissions_handler.is_user_allowed(invitee, inviter):
                logger.debug(
                    "Not allowing invite since local user (%s) did not allow the remote user (%s) in their permissions",
                    invitee,
                    inviter,
                )
                return errors.Codes.FORBIDDEN

        logger.debug(
            "Allowing invite since no other permission checks block. (%s -> %s)",
            invitee,
            inviter,
        )

        return NOT_SPAM

    async def get_all_room_ids(self) -> set[str]:
        """Retrieve all room IDS."""

        # There is an PRIMARY index on room_id
        def f(txn: LoggingTransaction) -> set[str]:
            sql = "SELECT room_id FROM rooms"
            txn.execute(sql)
            return {room_id for (room_id,) in txn.fetchall()}

        return await self.api._store.db_pool.runInteraction("get_rooms", f)

    @measure_func("get_timestamps_from_eligible_events_for_epa_room_purge")
    async def get_timestamps_from_eligible_events_for_epa_room_purge(
        self, room_id
    ) -> EpaRoomTimestampResults:
        """
        Retrieve and parse the room PRO members that left into a timestamp of the
        latest event, or the timestamp of the create event if there were no PRO members
        in the room

        Returns None when no events were found for a room
        """
        state_mapping: StateMap[EventBase] = (
            await self.api._storage_controllers.state.get_current_state(
                room_id,
                StateFilter.from_types(
                    [(EventTypes.Member, None), (EventTypes.Create, None)]
                ),
            )
        )

        leave_event_timestamps = set()
        invite_event_timestamps = set()
        create_event_ts = None

        for (state_type, state_key), event in state_mapping.items():
            if state_type == EventTypes.Create:
                create_event_ts = event.origin_server_ts

            elif state_type == EventTypes.Member:
                users_domain = UserID.from_string(state_key).domain
                if not await self.is_domain_insurance(users_domain):
                    if event.membership == Membership.LEAVE:
                        leave_event_timestamps.add(event.origin_server_ts)

                    elif event.membership == Membership.INVITE:
                        invite_event_timestamps.add(event.origin_server_ts)

        return EpaRoomTimestampResults(
            max(invite_event_timestamps) if invite_event_timestamps else None,
            max(leave_event_timestamps) if leave_event_timestamps else None,
            create_event_ts,
        )

    @measure_func("get_timestamp_of_last_eligible_activity_in_room")
    async def get_timestamp_of_last_eligible_activity_in_room(
        self, room_id: str
    ) -> int | None:
        """
        A two-staged approach to finding the last activity in a room.
        First, search a room for the last message(either encrypted or plaintext) or
        creation event timestamp in that room(which ever is found to be most recent).
        This approach works best for local only rooms or rooms that were initialized
        locally.

        Second, if none of those were found, it's likely this room is from a remote
        server and it's events are still outliers. Try to find membership events that
        do not have a matching state_key and sender, as these are most likely invite
        events. The possibility of them being a kick or ban is there, but unlikely as
        the previous stage would have caught the creation event instead

        Returns None when no events were found for a room
        """
        # Including a type doesn't guarantee that at least one of each is present in the result
        filter_json = {
            "types": [EventTypes.Message, EventTypes.Encrypted, EventTypes.Create]
        }

        events = await self._get_filtered_events_from_pagination(room_id, filter_json)

        collected_timestamps = {event_base.origin_server_ts for event_base in events}
        if collected_timestamps:
            return max(collected_timestamps)

        # So, maybe this is a remote invite situation. The pagination handler does not
        # pick up outliers and out-of-band memberships until they have been de-outliered
        # such as through backfill. Retrieve these manually. As an additional note: the
        # get_current_state() used in
        # get_timestamp_from_eligible_events_for_epa_room_purge() also does not appear
        # to get outlier events, neither did the partial_state variant.

        # This will either have the timestamp or be None if none was found
        return await self._last_explicit_membership_ts_in_room(room_id)

    async def _get_filtered_events_from_pagination(
        self, room_id: str, filter_json: dict[str, list[str]]
    ) -> list[EventBase]:
        """
        Providing a room_id and a json Filter will retrieve a list of(at most) 5
        EventBases from the newest Events in that room.

        Args:
            room_id: The room to target
            filter_json: The filter json as a dict to use

        Returns: list of EventBases

        """
        event_filter = Filter(self.api._hs, filter_json)

        from_token = (
            await self.api._hs.get_event_sources().get_current_token_for_pagination(
                room_id
            )
        )

        (
            events,
            next_key,
            _,
        ) = await self.api._store.paginate_room_events_by_topological_ordering(
            room_id=room_id,
            from_key=from_token.room_key,
            # When going backwards, to_key is not important
            to_key=None,
            direction=Direction.BACKWARDS,
            # With the filter below, 5 should be more than enough
            limit=5,
            event_filter=event_filter,
        )

        return events

    async def _last_explicit_membership_ts_in_room(self, room_id: str) -> int | None:
        """
        Retrieve the newest origin_server_ts for a membership event in a given room when
        the sender key does not match the state_key. This is going to be most likely
        an invite
        """
        # TODO: this function needs a better name
        # Unfortunately, there does not appear to be a nice helper to select a row when
        # the condition needs to have a `!=` compare clause
        sql = """
        SELECT MAX(origin_server_ts)
        FROM events
        WHERE room_id = ? AND type = 'm.room.member' AND sender != state_key
        """
        rows = cast(
            list[tuple[int]],
            await self.api._store.db_pool.execute(
                "_last_explicit_membership_ts_in_room", sql, room_id
            ),
        )

        for (origin_server_ts,) in rows:
            # There should only ever be one row, because of the MAX() in the sql
            return origin_server_ts

        # It is increasingly probable that this can never happen now. Local rooms will
        # have had a creation event to fallback on, and remote rooms will be private and
        # therefore require an invite. Leave it in for now, just in case a broken room
        # still can come along from a remote server. It is logged by room_scan().
        return None

    async def get_delete_tasks_by_room(self, room_id: str) -> list[ScheduledTask]:
        """Get scheduled or active delete tasks by room

        Args:
            room_id: room_id that is being targeted
        """
        # We specifically ignore "COMPLETED" and "FAILED" so they can be tried again
        # if the room scan found them still hanging around. This should only occur for
        # ridiculously complex rooms and should not be an issue in the gematik federation
        statuses = [TaskStatus.ACTIVE, TaskStatus.SCHEDULED]

        return await self.task_scheduler.get_tasks(
            actions=[SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME],
            resource_id=room_id,
            statuses=statuses,
        )

    async def kick_all_local_users(self, room_id: str) -> None:
        """
        First block the room from further interaction from this server, then kick all
        local members

        Args:
            room_id:
        """
        requester_user_id = (
            f"@_synapse-invite-checker:{self.api._hs.config.server.server_name}"
        )

        # Block the room, this prevents inadvertent rejoins and stray events/messages
        # from becoming part of the room after the fact but will still allow the user
        # leaving to occur.
        # NOTE: This call is idempotent, so even if some unforeseen error occurs that
        # causes the user to fail to leave the room, it will still be blocked and
        # calling this again won't unblock it or otherwise interfere.
        await self.api._store.block_room(room_id, requester_user_id)

        users = await self.api._store.get_local_users_related_to_room(room_id)
        for user_id, membership in users:
            # If the user is not in the room (or is banned), nothing to do.
            if membership not in (Membership.JOIN, Membership.INVITE, Membership.KNOCK):
                continue

            logger.info("Kicking %s from %s...", user_id, room_id)
            # Use the actual user as a puppet. Don't use the auto-kicker
            # user id above, as it won't pass room auth
            target_requester = create_requester(user_id, authenticated_entity=user_id)

            try:
                # Kick users from room
                (
                    _,
                    stream_id,
                ) = await self.api._hs.get_room_member_handler().update_membership(
                    requester=target_requester,
                    target=target_requester.user,
                    room_id=room_id,
                    action=Membership.LEAVE,
                    content={},
                    ratelimit=False,
                    require_consent=False,
                )

            except Exception:
                logger.exception(
                    "Failed to kick user '%s' from room %s", user_id, room_id
                )

    async def schedule_room_for_purge(self, room_id: str) -> None:
        """
        Schedules the deletion of a room from Synapse's database after kicking all users

        If the room has already been scheduled or is actively being deleted, do nothing.
        If the room was purged already in the past, but for some reason is still hanging
        around in the database, try it again
        """
        if len(await self.get_delete_tasks_by_room(room_id)) > 0:
            logger.warning("Purge already in progress or scheduled for %s", room_id)
            return

        shutdown_params = ShutdownRoomParams(
            new_room_user_id=None,
            new_room_name=None,
            message=None,
            requester_user_id=None,
            block=False,
            purge=True,  # <- to remove the room from the database
            force_purge=True,  # <- to force kick anyone else still in the room
        )

        delete_id = await self.task_scheduler.schedule_task(
            SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME,
            resource_id=room_id,
            params=shutdown_params,
            # Set the time to start to now, as we have already waited the requested time
            timestamp=self.api._hs.get_clock().time_msec(),
        )

        logger.info(
            "Scheduling shutdown and purge on room '%s' with delete_id '%s'",
            room_id,
            delete_id,
        )

    async def room_scan(self) -> None:
        """
        Scan all rooms for eligible conditions to shutdown and purge a room.
        """
        # Changed for version 0.4.2
        # To help mitigate the potential racing of a room scan during a room creation,
        # we will keep track that we already looked at this room. The next room scan
        # should pick up on this and then force the kick/purge if it still does not meet
        # the criteria. If this legitimately was a seriously broken room, the inactive
        # purge below will still completely remove it.

        all_room_ids = await self.get_all_room_ids()

        logger.debug("Detected %d total rooms", len(all_room_ids))

        server_notice_rooms = set()
        # On servers that don't have server notices set up, the mxid will be None
        if self.api._hs.config.servernotices.server_notices_mxid is not None:
            # This database call was chosen over `is_server_notice_room()` as batching
            # the data should be much more efficient than calling the latter for every
            # single room
            server_notice_rooms_list: list[RoomsForUser] = (
                await self.api._store.get_rooms_for_local_user_where_membership_is(
                    self.api._hs.config.servernotices.server_notices_mxid,
                    [Membership.JOIN],
                    [],
                )
            )
            # Simplify notice room list for fast lookup
            server_notice_rooms = {room.room_id for room in server_notice_rooms_list}

        if self.config.insured_room_scan_options.enabled:
            # grab a couple of references, this will be used frequently and helps make
            # the below conditions more readable.
            # XXX: should we do the same for 'current_time'?
            invites_grace_period_ms = (
                self.config.insured_room_scan_options.invites_grace_period_ms
            )
            grace_period_ms = self.config.insured_room_scan_options.grace_period_ms

            for room_id in all_room_ids:
                # Server notice rooms are exempt, don't want to necessarily hide that
                # a notice showed up before the user has a chance to see it.
                if room_id in server_notice_rooms:
                    continue

                # only shut down rooms that only have EPA hosts in them
                if await self.have_all_pro_hosts_left(room_id):
                    epa_room_results = await self.get_timestamps_from_eligible_events_for_epa_room_purge(
                        room_id
                    )
                    current_time = self.api._hs.get_clock().time_msec()

                    # Invites(if found) are ignored if the invites_grace_period is 0 or
                    # if the invites_grace_period + the actual timestamp of the invite
                    # are greater than now
                    if (
                        epa_room_results.last_invite_in_room is not None
                        and (
                            invites_grace_period_ms == 0
                            or (
                                epa_room_results.last_invite_in_room
                                + invites_grace_period_ms
                                > current_time
                            )
                        )
                        # Leaves(if found) are ignored if the grace_period + timestamp are
                        # greater than now
                        or (
                            epa_room_results.last_leave_in_room is not None
                            and epa_room_results.last_leave_in_room + grace_period_ms
                            > current_time
                        )
                        # There is not always a create event available, but if there is and
                        # its timestamp + grace_period are greater than now, then it is ignored
                        or epa_room_results.room_creation_ts is not None
                        and (
                            epa_room_results.room_creation_ts + grace_period_ms
                            > current_time
                        )
                    ):
                        pass

                    # In the unlikely event that all of the above did not happen, warn.
                    # Recall that we only get to this place if there are no Pro hosts in
                    # the room, so either an invite or a leave has to have occurred and
                    # in their absence we use the create event. So if there is no create
                    # event, then it is likely this room is broken somehow.
                    elif epa_room_results.room_creation_ts is None:
                        logger.warning(
                            "Not kicking users from room during ePA user room scan because it contained no events: %s",
                            room_id,
                        )
                    else:
                        await self.kick_all_local_users(room_id)

        if self.config.inactive_room_scan_options.enabled:
            for room_id in all_room_ids:
                last_eligible_event_ts = (
                    await self.get_timestamp_of_last_eligible_activity_in_room(room_id)
                )
                if last_eligible_event_ts is None:
                    logger.warning(
                        "A room was found that contained no events, skipping purge for: %s",
                        room_id,
                    )

                elif (
                    last_eligible_event_ts
                    + self.config.inactive_room_scan_options.grace_period_ms
                    <= self.api._hs.get_clock().time_msec()
                ):
                    await self.schedule_room_for_purge(room_id)

    async def have_all_pro_hosts_left(self, room_id: str) -> bool:
        """
        Retrieves all hosts that have a member in a given room, and filters them to
        decide if any of them are non-Epa servers.
        Args:
            room_id:

        Returns: bool representing if the hosts present are all non-Pro servers

        """
        hosts = await self.api._store.get_current_hosts_in_room(room_id=room_id)  # type: ignore[call-arg]
        for host in hosts:
            if not await self.is_domain_insurance(host):
                return False

        return True



# Copyright (C) 2025 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import json
from typing import Any
from unittest.mock import AsyncMock, patch

import yaml
from parameterized import parameterized_class
from synapse.api.errors import Codes, HttpResponseException
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from synapse_invite_checker.types import (
    LOCAL_SERVER_TEMPLATE,
    DefaultPermissionConfig,
    GroupName,
    PermissionConfig,
    PermissionConfigType,
)
from tests.base import FederatingModuleApiTestCase
from tests.test_utils import INSURANCE_DOMAIN_IN_LIST
from tests.unittest import TestCase


def strip_json_of_whitespace(test_json) -> str:
    """
    Canonicalize the JSON, so any given values are in the same place
    """
    return json.dumps(
        json.loads(test_json),
        # This strips whitespace from around the separators
        separators=(",", ":"),
        # Guarantee all keys are always in the same order
        sort_keys=True,
    )


def assert_test_json_matches_permissions(
    test_json: str, permissions: PermissionConfig
) -> None:
    """
    Test assert that stripping all the whitespace and sorting keys of the json yields
    the same json after it has passed through the PermissionConfig
    """
    # First, strip and order the existing test json from the test
    test_json_stripped = strip_json_of_whitespace(test_json)
    # Then, use the PermissionConfig to dump its data and juggle it back and forth into
    # a string that has also been sorted and stripped. They should match now. If they do
    # not, it means something that was not supposed to be included is contaminating the
    # json and the test should fail.
    # We purposely do not use the `model_dump_json` helper method from Pydantic here, as
    # that is not used in production and therefore would not be a truly sterile test(it
    # outputs a string of json, but does not sort and strip it for us anyway).
    assert test_json_stripped == strip_json_of_whitespace(
        json.dumps(permissions.dump())
    )


def assert_test_yaml_matches_json_dump(
    test_yaml: str, dumped_json: dict[str, Any], maybe_server_name: str | None
) -> None:
    """
    Test assert that given test yaml and resultant permissions match how we expect.

    Specifically, it is needed that the root level keys are in place and any sub-keys
    attached match. The data attached to those keys are not relevant as they are not
    used.

    """
    # convert yaml to dict
    converted_yaml: dict[str, Any] = yaml.safe_load(test_yaml)

    # If the local server was included by default with the template, we must account
    # for its existence here. Watch for it to be present, and look for the actual
    # server name in the dumped json
    for key in (
        "defaultSetting",
        "userExceptions",
        "serverExceptions",
        "groupExceptions",
    ):
        yaml_key_found = key in converted_yaml
        dj_key_found = key in dumped_json
        assert yaml_key_found == dj_key_found
        if yaml_key_found:
            for sub_key in converted_yaml[key]:
                if sub_key == LOCAL_SERVER_TEMPLATE:
                    assert maybe_server_name in dumped_json["serverExceptions"]
                    assert LOCAL_SERVER_TEMPLATE not in dumped_json["serverExceptions"]
                else:
                    assert sub_key in dumped_json[key]


class PermissionConfigTest(TestCase):
    @staticmethod
    def test_model_validate_permissions_default() -> None:
        test_json = '{"defaultSetting": "allow all"}'
        test_permission_object = PermissionConfig.model_validate_json(test_json)

        assert test_permission_object.is_allow_all()
        assert not test_permission_object.is_group_excepted(GroupName.isInsuredPerson)

        assert test_permission_object.is_mxid_allowed_to_contact(
            "@bob:example.com", is_mxid_epa=False
        )
        assert_test_json_matches_permissions(test_json, test_permission_object)

        test_json = '{"defaultSetting": "block all"}'
        test_permission_object = PermissionConfig.model_validate_json(test_json)

        assert not test_permission_object.is_allow_all()
        assert not test_permission_object.is_group_excepted(GroupName.isInsuredPerson)

        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@bob:example.com", is_mxid_epa=False
        )
        assert_test_json_matches_permissions(test_json, test_permission_object)

    def test_model_validate_permissions_complete(self) -> None:
        """
        Test complete forms with both "block all" and "allow all" behaviors. These both
        test with JSON and with a Dict. The both tests are relevant in that the
        PermissionConfig defines additional fields that if unused won't be None, and can
        not be in later JSON used by the account_data system.
        """
        # block all section
        test_json = """
        {
            "defaultSetting": "block all",
            "serverExceptions":
                {
                    "power.rangers": {}
                },
            "groupExceptions":
                [{
                    "groupName": "isInsuredPerson"
                }],
            "userExceptions":
                {
                    "@david:hassel.hoff": {}
                }
        }
        """
        test_permission_object = PermissionConfig.model_validate_json(test_json)

        assert "power.rangers" in test_permission_object.serverExceptions
        assert "@david:hassel.hoff" in test_permission_object.userExceptions

        assert not test_permission_object.is_allow_all()
        assert test_permission_object.is_group_excepted(GroupName.isInsuredPerson)
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@david:hassel.hoff", is_mxid_epa=False
        )
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@billy:power.rangers", is_mxid_epa=False
        )
        # This looks strange, but recall that the defaultSetting is "block all" so it is correct
        assert test_permission_object.is_mxid_allowed_to_contact(
            f"@patient:{INSURANCE_DOMAIN_IN_LIST}", is_mxid_epa=True
        )

        assert_test_json_matches_permissions(test_json, test_permission_object)

        test_dict = {
            "defaultSetting": "block all",
            "serverExceptions": {"power.rangers": {}},
            "groupExceptions": [{"groupName": "isInsuredPerson"}],
            "userExceptions": {"@david:hassel.hoff": {}},
        }

        test_permission_object = PermissionConfig.model_validate(test_dict)

        assert "power.rangers" in test_permission_object.serverExceptions
        assert "@david:hassel.hoff" in test_permission_object.userExceptions
        assert test_dict == test_permission_object.dump()

        assert not test_permission_object.is_allow_all()
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@david:hassel.hoff", is_mxid_epa=False
        )
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@billy:power.rangers", is_mxid_epa=False
        )
        assert test_permission_object.is_mxid_allowed_to_contact(
            f"@patient:{INSURANCE_DOMAIN_IN_LIST}", is_mxid_epa=True
        )
        assert test_permission_object.is_group_excepted(GroupName.isInsuredPerson)

        # allow all section
        test_json = """
                {
                    "defaultSetting": "allow all",
                    "serverExceptions":
                        {
                            "power.rangers": {}
                        },
                    "groupExceptions":
                        [{
                            "groupName": "isInsuredPerson"
                        }],
                    "userExceptions":
                        {
                            "@david:hassel.hoff": {}
                        }
                }
                """
        test_permission_object = PermissionConfig.model_validate_json(test_json)

        assert "power.rangers" in test_permission_object.serverExceptions
        assert "@david:hassel.hoff" in test_permission_object.userExceptions

        assert test_permission_object.is_allow_all()
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@david:hassel.hoff", is_mxid_epa=False
        )
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@billy:power.rangers", is_mxid_epa=False
        )
        assert not test_permission_object.is_mxid_allowed_to_contact(
            f"@patient:{INSURANCE_DOMAIN_IN_LIST}", is_mxid_epa=True
        )

        assert_test_json_matches_permissions(test_json, test_permission_object)

        test_dict = {
            "defaultSetting": "allow all",
            "serverExceptions": {"power.rangers": {}},
            "groupExceptions": [{"groupName": "isInsuredPerson"}],
            "userExceptions": {"@david:hassel.hoff": {}},
        }

        test_permission_object = PermissionConfig.model_validate(test_dict)

        assert "power.rangers" in test_permission_object.serverExceptions
        assert "@david:hassel.hoff" in test_permission_object.userExceptions
        assert test_dict == test_permission_object.dump()

        assert test_permission_object.is_allow_all()
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@david:hassel.hoff", is_mxid_epa=False
        )
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@billy:power.rangers", is_mxid_epa=False
        )
        assert not test_permission_object.is_mxid_allowed_to_contact(
            f"@patient:{INSURANCE_DOMAIN_IN_LIST}", is_mxid_epa=True
        )

    def test_model_validate_permissions_scenarios(self) -> None:
        """
        Test both with "block_all" and "allow_all" in various scenarios
        """
        # scenarios
        # 1. a doctor(HBA) has allowed all contact but restricted insured actors
        # 2. an organization user has allowed all contact, but doesn't want to hear from
        #    a pharmacy domain that continuously misreads orders
        # 3. a patient has allowed all, but doesn't want to talk to that weird doctor
        # 4. a doctor has allowed all except insured and forgot he had blocked a specific patient

        # scenario 1
        test_json = """
        {
            "defaultSetting": "allow all",
            "groupExceptions":
                [{
                    "groupName": "isInsuredPerson"
                }]
        }
        """
        test_permission_object = PermissionConfig.model_validate_json(test_json)

        assert test_permission_object.is_allow_all()
        # insured are denied
        assert not test_permission_object.is_mxid_allowed_to_contact(
            f"@patient:{INSURANCE_DOMAIN_IN_LIST}", is_mxid_epa=True
        )
        # everyone else is ok
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@billy:power.rangers", is_mxid_epa=False
        )

        assert_test_json_matches_permissions(test_json, test_permission_object)

        # scenario 2
        test_json = """
        {
            "defaultSetting": "allow all",
            "serverExceptions":
                {
                    "pharmacy.com": {}
                }
        }
        """
        test_permission_object = PermissionConfig.model_validate_json(test_json)

        assert test_permission_object.is_allow_all()
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@needsglasses:pharmacy.com", is_mxid_epa=False
        )
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@twentytwentyvision:otherpharmacy.com", is_mxid_epa=False
        )

        assert_test_json_matches_permissions(test_json, test_permission_object)

        # scenario 3
        test_json = """
        {
            "defaultSetting": "allow all",
            "userExceptions":
                {
                    "@badbreath:doctors.edu": {}
                }
        }
        """
        test_permission_object = PermissionConfig.model_validate_json(test_json)

        assert test_permission_object.is_allow_all()
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@badbreath:doctors.edu", is_mxid_epa=False
        )
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@mothertheresa:doctors.edu", is_mxid_epa=False
        )

        assert_test_json_matches_permissions(test_json, test_permission_object)

        # scenario 4
        test_json = """
        {
            "defaultSetting": "allow all",
            "userExceptions":
                {
                    "@patient:insured.com": {}
                },
            "groupExceptions":
                [{
                    "groupName": "isInsuredPerson"
                }]
        }
        """
        test_permission_object = PermissionConfig.model_validate_json(test_json)

        assert test_permission_object.is_allow_all()
        # Even though this patient permission exists, they are still blocked
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@patient:insured.com", is_mxid_epa=True
        )
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@ghandi:insured.com", is_mxid_epa=True
        )

        assert_test_json_matches_permissions(test_json, test_permission_object)


class DefaultPermissionConfigTest(TestCase):
    """
    This set of cases differs from PermissionConfigTest above in that it has to also
    accommodate the local server name as a template that can be included into the
    serverExceptions section
    """

    server_name = "server_name.com"

    def test_model_validate_permissions_default(self) -> None:
        test_yaml = """
        defaultSetting: allow all
        """
        converted_yaml = yaml.safe_load(test_yaml)
        test_permission_object = DefaultPermissionConfig.model_validate(converted_yaml)

        assert test_permission_object.defaultSetting is not None
        assert test_permission_object.defaultSetting == "allow all"

        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), None
        )

        # Since the mode is "allow all", this will be ignored
        test_permission_object.maybe_update_server_exceptions(self.server_name)
        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), None
        )

        test_yaml = """
        defaultSetting: block all
        """
        converted_yaml = yaml.safe_load(test_yaml)
        test_permission_object = DefaultPermissionConfig.model_validate(converted_yaml)

        assert test_permission_object.defaultSetting is not None
        assert test_permission_object.defaultSetting == "block all"

        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), None
        )

        # Also test that the default of "block all" still tests as expected when
        # the local server template object is missing
        test_permission_object.maybe_update_server_exceptions(self.server_name)
        assert test_permission_object.defaultSetting is not None
        assert test_permission_object.defaultSetting == "block all"

        assert (
            self.server_name not in test_permission_object.serverExceptions
        ), f"No server should be an exception since no {LOCAL_SERVER_TEMPLATE} was included"
        assert isinstance(test_permission_object.groupExceptions, list)

        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), self.server_name
        )

        test_yaml = """
        defaultSetting: block all
        serverExceptions:
          "@LOCAL_SERVER@":
        """
        converted_yaml = yaml.safe_load(test_yaml)
        test_permission_object = DefaultPermissionConfig.model_validate(converted_yaml)

        # Also test that the default of "block all" still tests as expected when
        # establishing the local server as the exemption
        test_permission_object.maybe_update_server_exceptions(self.server_name)
        assert test_permission_object.defaultSetting is not None
        assert test_permission_object.defaultSetting == "block all"
        assert isinstance(test_permission_object.userExceptions, dict)
        assert isinstance(test_permission_object.serverExceptions, dict)
        assert (
            self.server_name in test_permission_object.serverExceptions
        ), f"server should be an exception since {LOCAL_SERVER_TEMPLATE} was included"
        assert isinstance(test_permission_object.groupExceptions, list)

        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), self.server_name
        )

    def test_model_validate_permissions_complete(self) -> None:
        """
        Test complete forms with both "block all" and "allow all" behaviors. These both
        test with JSON and with a Dict. The both tests are relevant in that the
        PermissionConfig defines additional fields that if unused won't be None, and can
        not be in later JSON used by the account_data system.
        """
        # block all section
        test_yaml = """
        defaultSetting: block all
        serverExceptions:
          power.rangers:
        groupExceptions:
        - groupName: isInsuredPerson
        userExceptions:
          "@david:hassel.hoff": {}
        """
        converted_yaml = yaml.safe_load(test_yaml)

        test_permission_object = DefaultPermissionConfig.model_validate(converted_yaml)

        assert "power.rangers" in test_permission_object.serverExceptions
        assert "@david:hassel.hoff" in test_permission_object.userExceptions

        # Strictly speaking, these methods aren't used by DefaultPermissionConfig, but
        # they came along for the ride so may as well use them
        assert not test_permission_object.is_allow_all()
        assert test_permission_object.is_group_excepted(GroupName.isInsuredPerson)
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@david:hassel.hoff", is_mxid_epa=False
        )
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@billy:power.rangers", is_mxid_epa=False
        )
        assert test_permission_object.is_mxid_allowed_to_contact(
            f"@patient:{INSURANCE_DOMAIN_IN_LIST}", is_mxid_epa=True
        )

        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), None
        )

        # allow all section
        test_yaml = """
            defaultSetting: allow all
            serverExceptions:
              power.rangers:
            groupExceptions:
            - groupName: isInsuredPerson
            userExceptions:
              "@david:hassel.hoff": {}
        """
        converted_yaml = yaml.safe_load(test_yaml)

        test_permission_object = DefaultPermissionConfig.model_validate(converted_yaml)

        assert "power.rangers" in test_permission_object.serverExceptions
        assert "@david:hassel.hoff" in test_permission_object.userExceptions

        assert test_permission_object.is_allow_all()
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@david:hassel.hoff", is_mxid_epa=False
        )
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@billy:power.rangers", is_mxid_epa=False
        )
        assert not test_permission_object.is_mxid_allowed_to_contact(
            f"@patient:{INSURANCE_DOMAIN_IN_LIST}", is_mxid_epa=True
        )
        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), None
        )

    def test_model_validate_permissions_scenarios(self) -> None:
        """
        Test both with "block_all" and "allow_all" in various scenarios
        """
        # scenarios
        # 1. a doctor(HBA) has allowed all contact but restricted insured actors
        # 2. an organization user has allowed all contact, but doesn't want to hear from
        #    a pharmacy domain that continuously misreads orders
        # 3. a patient has allowed all, but doesn't want to talk to that weird doctor
        # 4. a doctor has allowed all except insured and forgot he had blocked a
        #    specific patient(which still blocks)

        # scenario 1
        test_yaml = """
        defaultSetting: allow all
        groupExceptions:
        - groupName: isInsuredPerson
        """

        converted_yaml = yaml.safe_load(test_yaml)
        test_permission_object = DefaultPermissionConfig.model_validate(converted_yaml)

        assert test_permission_object.is_allow_all()
        # insured are denied
        assert not test_permission_object.is_mxid_allowed_to_contact(
            f"@patient:{INSURANCE_DOMAIN_IN_LIST}", is_mxid_epa=True
        )
        # everyone else is ok
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@billy:power.rangers", is_mxid_epa=False
        )

        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), None
        )

        # scenario 2
        test_yaml = """
        defaultSetting: allow all
        serverExceptions:
          pharmacy.com:
        """

        converted_yaml = yaml.safe_load(test_yaml)
        test_permission_object = DefaultPermissionConfig.model_validate(converted_yaml)

        assert test_permission_object.is_allow_all()
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@needsglasses:pharmacy.com", is_mxid_epa=False
        )
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@twentytwentyvision:otherpharmacy.com", is_mxid_epa=False
        )

        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), None
        )

        # scenario 3
        test_yaml = """
        defaultSetting: allow all
        userExceptions:
          "@badbreath:doctors.edu":
        """
        converted_yaml = yaml.safe_load(test_yaml)
        test_permission_object = DefaultPermissionConfig.model_validate(converted_yaml)

        assert test_permission_object.is_allow_all()
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@badbreath:doctors.edu", is_mxid_epa=False
        )
        assert test_permission_object.is_mxid_allowed_to_contact(
            "@mothertheresa:doctors.edu", is_mxid_epa=False
        )

        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), None
        )

        # scenario 4
        test_yaml = """
        defaultSetting: allow all
        userExceptions:
          "@patient:insured.com":
        groupExceptions:
        - groupName: isInsuredPerson
        """

        converted_yaml = yaml.safe_load(test_yaml)
        test_permission_object = DefaultPermissionConfig.model_validate(converted_yaml)

        assert test_permission_object.is_allow_all()
        # Even though this patient permission exists, they are still blocked
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@patient:insured.com", is_mxid_epa=True
        )
        assert not test_permission_object.is_mxid_allowed_to_contact(
            "@ghandi:insured.com", is_mxid_epa=True
        )

        assert_test_yaml_matches_json_dump(
            test_yaml, test_permission_object.dump(), None
        )


class LoginGeneratesDefaultPermissionsTestCase(FederatingModuleApiTestCase):
    """
    Test logins generate initial set of permissions for contacting other users
    """

    # This test case does not use any federation features
    # By default, we are SERVER_NAME_FROM_LIST
    # server_name_for_this_server = "tim.test.gematik.de"
    # This test case will model being an PRO server on the federation list

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)

        self.user_a = self.register_user("a", "password")
        self.user_b = self.register_user("b", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        default_perms = {
            "defaultSetting": "allow all",
            "groupExceptions": [{"groupName": "isInsuredPerson"}],
        }
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": default_perms}
        )

        return conf

    def test_login(self) -> None:
        """Test that login will populate the permission structure"""
        # We only pick on user "a" in this test
        config_type = self.inv_checker.permissions_handler.permission_event_type()

        existing_data = self.get_success(
            self.module_api.account_data_manager.get_global(
                self.user_a, config_type.value
            )
        )

        assert existing_data is None, "Initial permission data should not exist"

        # Login should trigger the initial permission configuration
        self.login("a", "password")

        existing_data = self.get_success(
            self.module_api.account_data_manager.get_global(
                self.user_a, config_type.value
            )
        )
        assert existing_data is not None, "Data should have been populated"

    def test_invite_before_login(self) -> None:
        """
        Test that an invite before user's first login will behave as default
        permissions expect
        """
        # We only pick on user "b" in this test
        config_type = self.inv_checker.permissions_handler.permission_event_type()

        existing_data = self.get_success(
            self.module_api.account_data_manager.get_global(
                self.user_b, config_type.value
            )
        )

        assert existing_data is None, "Initial permission data should not exist"
        # User "b" has already registered as a user, but never signed in. What happens
        # if an invite comes in that is expected to be denied? Since we set this up as
        # "allow all" but denied insured persons, this should fail
        result = self.get_success_or_raise(
            self.inv_checker.user_may_invite(
                f"@rando-from:{INSURANCE_DOMAIN_IN_LIST}",
                self.user_b,
                f"!example-room:{INSURANCE_DOMAIN_IN_LIST}",
            )
        )
        assert result == Codes.FORBIDDEN, "Invite should have been FORBIDDEN"


@parameterized_class(
    [
        {
            "overridden_server_mode": "epa",
            "expected_perm_type": PermissionConfigType.EPA_ACCOUNT_DATA_TYPE,
        },
        {
            "overridden_server_mode": "pro",
            "expected_perm_type": PermissionConfigType.PRO_ACCOUNT_DATA_TYPE,
        },
    ]
)
class FederationListFailureScenarioTestCase(FederatingModuleApiTestCase):
    """
    Test two scenarios here:
    1. That a missing federation list does not tank the server and cause an exception
        during startup, and
    2. That login will not trip up on a missing federation list anymore. Strictly
        speaking, the behavior where the federation list was used to decide what
        permission object type no longer exists so this just verifies it will not break again.
    """

    # I think in the future we may move this test case somewhere more appropriate and
    # extend it for checking other behaviors around a missing federation list

    overridden_server_mode = None
    expected_perm_type = None
    # This test case does not use any federation features
    # NOTE: because this test does not actually check permissions or invites, it does
    #  not actually set the server name as something appropriate. In theory, this is
    #  wrong, but the only outward facing sign would be the startup warning in the logs

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Need to bork this one for the test
        cls._patcher1.stop()
        cls._patcher1 = patch(
            "synapse_invite_checker.InviteChecker._raw_federation_list_fetch",
            new=AsyncMock(side_effect=HttpResponseException(404, "Not Found", b"")),
        )

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        # Create the user, but do not log in until the test as that is the behavior that
        # is being tested
        self.user_c = self.register_user("c", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"
        conf["modules"][0].setdefault("config", {}).update(
            {"tim-type": self.overridden_server_mode}
        )
        default_perms = {
            "defaultSetting": "allow all",
        }
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": default_perms}
        )

        return conf

    def test_login_with_no_fed_list(self) -> None:
        """Test that login will populate the permission structure even with no fed list"""
        # We only pick on user "c" in this test
        config_type = self.inv_checker.permissions_handler.permission_event_type()

        assert config_type == self.expected_perm_type, "Wrong permission type"

        existing_data = self.get_success(
            self.module_api.account_data_manager.get_global(
                self.user_c, config_type.value
            )
        )

        assert existing_data is None, "Initial permission data should not exist"

        self.login("c", "password")

        existing_data = self.get_success(
            self.module_api.account_data_manager.get_global(
                self.user_c, config_type.value
            )
        )
        assert existing_data is not None, "Data should have been populated"


class PermissionsAccountDataValidationEdgeTestCases(FederatingModuleApiTestCase):
    """
    Test various potential issues that may occur from permissions being against schema
    """

    # This test case does not use any federation features
    # By default, we are SERVER_NAME_FROM_LIST
    # server_name_for_this_server = "tim.test.gematik.de"
    # This test case will model being an PRO server on the federation list

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)

        self.user_a = self.register_user("a", "password")
        self.user_b = self.register_user("b", "password")
        self.user_c = self.register_user("c", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        default_perms = {
            "defaultSetting": "allow all",
            # may need to lose this when testing against EPA
            "groupExceptions": [{"groupName": "isInsuredPerson"}],
        }
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": default_perms}
        )

        return conf

    def test_reading_permissions_schema_violation(self) -> None:
        """Test that existing permissions objects that don't meet the specification get reset"""
        # We only pick on user "a" in this test
        config_type = self.inv_checker.permissions_handler.permission_event_type()

        existing_data = self.get_success(
            self.module_api.account_data_manager.get_global(
                self.user_a, config_type.value
            )
        )

        assert existing_data is None, "Initial permission data should not exist"

        # As a simple example, what happens if we make groupExceptions not a List?
        test_json = """
        {
            "defaultSetting": "block all",
            "serverExceptions":
                {
                    "power.rangers": {}
                },
            "groupExceptions":
                {
                    "groupName": "isInsuredPerson"
                },
            "userExceptions":
                {
                    "@david:hassel.hoff": {}
                }
        }
        """
        test_json_converted = json.loads(test_json)
        self.get_success_or_raise(
            self.module_api.account_data_manager.put_global(
                self.user_a, config_type.value, test_json_converted
            )
        )
        perms = self.get_success_or_raise(
            self.inv_checker.permissions_handler.get_permissions(self.user_a)
        )

        assert (
            perms == self.inv_checker.permissions_handler.default_perms
        ), "permissions should be default after reset"

    def test_reading_permissions_schema_empty(self) -> None:
        """Test that permissions objects that are empty are reset"""
        # We only pick on user "c" in this test
        config_type = self.inv_checker.permissions_handler.permission_event_type()

        existing_data = self.get_success(
            self.module_api.account_data_manager.get_global(
                self.user_c, config_type.value
            )
        )

        assert existing_data is None, "Initial permission data should not exist"

        # As a simple example, what happens if we make the entire structure empty?
        test_json = "{}"
        test_json_converted = json.loads(test_json)
        self.get_success_or_raise(
            self.module_api.account_data_manager.put_global(
                self.user_c, config_type.value, test_json_converted
            )
        )
        perms = self.get_success_or_raise(
            self.inv_checker.permissions_handler.get_permissions(self.user_c)
        )
        assert (
            perms == self.inv_checker.permissions_handler.default_perms
        ), "permissions should be default after reset"



# Copyright (C) 2025 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# mypy: disable-error-code=call-arg
from typing import Any

import pytest
from parameterized import parameterized
from synapse.api.constants import EventTypes, JoinRules, Membership
from synapse.api.errors import AuthError
from synapse.api.room_versions import KNOWN_ROOM_VERSIONS
from synapse.handlers.pagination import SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME
from synapse.server import HomeServer
from synapse.types import TaskStatus, UserID
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from tests.base import FederatingModuleApiTestCase
from tests.test_utils import (
    DOMAIN_IN_LIST,
    INSURANCE_DOMAIN_IN_LIST,
    INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL,
    SERVER_NAME_FROM_LIST,
    event_injection,
)


class InsuredOnlyRoomScanTaskTestCase(FederatingModuleApiTestCase):
    """
    Test that insured only room scans are done, and required room kicks are done
    """

    # This test case will model being an EPA server on the federation list
    remote_pro_user = f"@mxid:{DOMAIN_IN_LIST}"
    remote_pro_user_2 = f"@a:{SERVER_NAME_FROM_LIST}"
    remote_epa_user = f"@alice:{INSURANCE_DOMAIN_IN_LIST}"
    remote_epa_user_2 = f"@bob:{INSURANCE_DOMAIN_IN_LIST}"
    # Our server name
    server_name_for_this_server = INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL
    # The default "fake" remote server name that has its server signing keys auto-injected
    OTHER_SERVER_NAME = DOMAIN_IN_LIST

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "epa"})
        conf["modules"][0].setdefault("config", {}).update(
            {"room_scan_run_interval": "1h"}
        )
        conf["modules"][0].setdefault("config", {}).update(
            {
                "insured_only_room_scan": {
                    "enabled": True,
                    "grace_period": "6h",
                    "invites_grace_period": "7h",
                },
                # Enabled to test broken rooms being ignored(but later purged correctly)
                # We don't worry about the grace_period as it will not apply
                "inactive_room_scan": {"enabled": True},
            }
        )
        conf["server_notices"] = {"system_mxid_localpart": "server", "auto_join": True}

        return conf

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.task_scheduler = self.hs.get_task_scheduler()

        self.user_d = self.register_user("d", "password")
        # Need the full UserID in a few places for sending a message into the room
        self.user_d_id = UserID.from_string(self.user_d)
        self.user_e = self.register_user("e", "password")
        self.login("d", "password")
        # Need this access token for the invite helper below
        self.access_token_e = self.login("e", "password")

        # OTHER_SERVER_NAME already has it's signing key injected into our database so
        # our server doesn't have to make that request. Add the other servers we will be
        # using as well
        self.inject_servers_signing_key(INSURANCE_DOMAIN_IN_LIST)
        self.inject_servers_signing_key(SERVER_NAME_FROM_LIST)

    @parameterized.expand(
        [
            ("pro_join_and_leave", True, True),
            ("pro_never_join", False, True),
            ("pro_never_invited_or_joined", False, False),
        ]
    )
    def test_room_scan_detects_epa_rooms(
        self, pro_activity: str, pro_join: bool, pro_invited: bool
    ) -> None:
        """
        Test that a room is handled as appropriately with a single EPA user and a single
        PRO user. Also test if the PRO user never joined/left(to test that 'maybe broken'
        rooms) and rooms with no invites at all.
        """
        # Make a room...
        room_id = self.create_local_room(self.user_d, [], is_public=False)
        assert room_id is not None, "Room should be created"

        # ...then (maybe) invite the doctor
        if pro_invited:
            self.helper.invite(
                room_id,
                self.user_d,
                self.remote_pro_user,
                tok=self.map_user_id_to_token[self.user_d],
            )
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        # Needs to be either None or False
        assert not is_room_blocked, "Room should not be blocked yet(try 1)"

        # doctor joins
        if pro_join:
            self.send_join(self.remote_pro_user, room_id)

        # Send a junk hex message into the room, like a sentinel
        self.create_and_send_event(room_id, self.user_d_id)

        # doctor leaves room
        if pro_join:
            self.send_leave(self.remote_pro_user, room_id)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 2)"

        # The config gives 6 hours for leaves and 7 hours for joins. Depending on the test
        # we fast-forward an hour short of which one is expected.
        if pro_activity == "pro_never_join":
            # The 'pro_never_join' is specifically a test for an invite without a join
            self.reactor.advance(6 * 60 * 60)
        else:
            self.reactor.advance(5 * 60 * 60)

        # Room should still exist
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 3)"

        # The TaskScheduler has a heartbeat of 1 minute, give it that much
        self.reactor.advance(60 * 60)

        # Now the room should be gone
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert is_room_blocked, "Room should be blocked now(try 4)"

        # Send a junk hex message into the room, like a sentinel
        # Inside EventCreationHandler.handle_new_client_event(), this raises as an
        # AuthError(which is a subclass of SynapseError). It appears to be annotated with a 403 as well
        with pytest.raises(AuthError):
            self.create_and_send_event(room_id, self.user_d_id)

    def test_room_scan_skips_incomplete_epa_rooms(self) -> None:
        """
        Test that a partially formed room does not break the room scanner, but instead
        is skipped and will not be purged(if enabled). Remember, rooms scans run every
        hour by default in this TestCase
        """
        # Make the bare minimum of a room, which is basically just a row in the `rooms` table
        room_version_id = self.hs.config.server.default_room_version.identifier

        room_version = KNOWN_ROOM_VERSIONS.get(room_version_id)
        assert room_version
        room_id = self.get_success_or_raise(
            self.hs.get_room_creation_handler()._generate_and_create_room_id(
                self.user_d, is_public=False, room_version=room_version
            )
        )
        assert room_id is not None, "Room should be partially created"

        # Even with an incomplete room, blocking should still work as expected. It will
        # be blocked as part of the insured kicking procedure. If it is not blocked, then
        # it is not scheduled for having members kicked out.
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        # Needs to be either None or False
        assert not is_room_blocked, "Room should not be blocked yet(try 1)"

        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should still exist(try 1)"

        # The config set up says it gets 1 hour for room scan frequency
        self.reactor.advance(60 * 60)

        # Normally, the room scan process would pick up the room here. But, we are
        # skipping the insured members kicking procedure, so it gets a second run
        # through before it is eligible for purge
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked(try 2)"

        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should still exist(try 2)"

        # try again, same setup
        self.reactor.advance(60 * 60)

        # The TaskScheduler has a heartbeat of 1 minute, give it that much
        self.reactor.advance(60)

        # The room will not register as blocked, since that only runs when the kicking
        # procedure is triggered
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked(try 3)"

        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should exist(try 3)"

    def test_room_scan_ignores_server_notices_rooms(self) -> None:
        """
        Test that a room is ignored when it is a server notices room
        """
        event_base = self.get_success_or_raise(
            self.hs.get_server_notices_manager().send_notice(
                self.user_d, {"body": "Server Notice message", "msgtype": "m.text"}
            )
        )
        room_id = event_base.room_id
        # Retrieving the room_id is a sign that the room was created, the user was
        # invited, and the message was sent
        assert room_id, "Server notices room should have been found"

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        # Needs to be either None or False
        assert not is_room_blocked, "Room should not be blocked yet(try 1)"

        self.reactor.advance(5 * 60 * 60)

        # Room should still exist
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 3)"

        # One more hour should be the 6 hours from settings
        self.reactor.advance(60 * 60)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should still exist"

        # Message should succeed, showing the room has not yet been left
        self.get_success_or_raise(
            self.hs.get_server_notices_manager().send_notice(
                self.user_d, {"body": "Server Notice message #2", "msgtype": "m.text"}
            )
        )

    def test_room_scan_detects_only_epa_rooms_with_multiple_hosts(self) -> None:
        """
        Test that a room is not deleted until the last PRO user leaves a room
        """
        # Make a room and invite the doctor
        room_id = self.create_local_room(
            self.user_d, [self.remote_pro_user], is_public=False
        )
        assert room_id is not None

        # Will need this later to include the correct stripped state with an invite
        room_version = self.get_success(self.store.get_room_version(room_id))

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        # Needs to be either None or False
        assert not is_room_blocked, "Room should not be blocked yet(try 1)"

        # doctor joins
        self.send_join(self.remote_pro_user, room_id)

        # doctor invites 2 more patients, because the first patient isn't allowed. One
        # is remote and the other is from this EPA server
        self.get_success_or_raise(
            event_injection.inject_member_event(
                self.hs,
                room_id,
                self.remote_pro_user,
                Membership.INVITE,
                target=self.remote_epa_user,
            )
        )

        # other patient joins
        self.send_join(self.remote_epa_user, room_id)

        self.get_success_or_raise(
            event_injection.inject_member_event(
                self.hs,
                room_id,
                self.remote_pro_user,
                Membership.INVITE,
                target=self.user_e,
                unsigned={
                    "invite_room_state": [
                        {
                            "type": EventTypes.Create,
                            "sender": self.user_d_id.to_string(),
                            # starting in room version 11, the 'sender' field is used
                            # instead. For the moment, we are limited to versions "9"
                            # and "10"
                            "content": {
                                "creator": self.user_d_id.to_string(),
                                "room_version": room_version.identifier,
                            },
                            "state_key": "",
                        },
                        {
                            "type": EventTypes.JoinRules,
                            "sender": self.user_d_id.to_string(),
                            "content": {"join_rule": JoinRules.PRIVATE},
                            "state_key": "",
                        },
                    ]
                },
            )
        )

        self.helper.join(room_id, self.user_e, tok=self.access_token_e)

        # doctor invites another doctor. To properly test joining, the invite from the
        # doctor needs to include the 'stripped_state' like bits in the invite itself
        self.get_success_or_raise(
            event_injection.inject_member_event(
                self.hs,
                room_id,
                self.remote_pro_user,
                Membership.INVITE,
                target=self.remote_pro_user_2,
            )
        )

        # other doctor joins
        self.send_join(self.remote_pro_user_2, room_id)

        # Send a junk hex message into the room, like a sentinel
        self.create_and_send_event(room_id, self.user_d_id)

        # They all just found out a friend of the remote patient may have more info
        self.get_success_or_raise(
            event_injection.inject_member_event(
                self.hs,
                room_id,
                self.remote_pro_user,
                Membership.INVITE,
                target=self.remote_epa_user_2,
            )
        )

        # other patient joins
        self.send_join(self.remote_epa_user_2, room_id)

        # Original patient says "thanks you can go now"
        self.create_and_send_event(room_id, self.user_d_id)

        # friend of friend of patient leaves
        self.send_leave(self.remote_epa_user_2, room_id)

        # doctor 1 leaves room
        self.send_leave(self.remote_pro_user, room_id)

        # The other local insured leaves the room
        self.helper.leave(room_id, self.user_e, tok=self.access_token_e)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 2)"

        self.reactor.advance(5 * 60 * 60)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 3)"

        # Normally, this would trigger the auto-kicker. But the doctor hasn't left yet
        self.reactor.advance(60 * 60)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 4)"

        # doctor 2 leaves
        self.send_leave(self.remote_pro_user_2, room_id)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 5)"

        self.reactor.advance(5 * 60 * 60)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 6)"

        # One more hour should be the 6 hours from settings
        self.reactor.advance(60 * 60)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert is_room_blocked, "Room should be blocked now"

        # Inside EventCreationHandler.handle_new_client_event(), this raises as an
        # AuthError(which is a subclass of SynapseError). It appears to be annotated with a 403 as well
        with pytest.raises(AuthError):
            self.create_and_send_event(room_id, self.user_d_id)

    def test_room_scan_detects_invites_as_participation(self) -> None:
        """
        Test that a room is not detected as inactive if one of two doctors never show up
        """
        # Make a room and invite the doctor
        room_id = self.create_local_room(self.user_d, [], is_public=False)
        assert room_id is not None

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        # Needs to be either None or False
        assert not is_room_blocked, "Room should not be blocked yet(try 1)"

        # Invite both pro users
        for remote_user in [self.remote_pro_user, self.remote_pro_user_2]:
            self.helper.invite(
                room_id,
                src=self.user_d,
                targ=remote_user,
                tok=self.map_user_id_to_token[self.user_d],
            )

        # doctor #1 joins
        self.send_join(self.remote_pro_user, room_id)

        # Send a junk hex message into the room, like a sentinel
        self.create_and_send_event(room_id, self.user_d_id)

        # doctor 1 leaves room, still have that pending invite for doctor #2
        self.send_leave(self.remote_pro_user, room_id)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 2)"

        self.reactor.advance(5 * 60 * 60)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 3)"

        # Normally, this would trigger the auto-kicker for leaves. But there is the
        # pending invite still waiting, which is configured to 7 hours and not 6
        self.reactor.advance(60 * 60)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 4)"

        # One more hour should be the 7 hours from settings
        self.reactor.advance(60 * 60)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert is_room_blocked, "Room should be blocked now"

        # Inside EventCreationHandler.handle_new_client_event(), this raises as an
        # AuthError(which is a subclass of SynapseError). It appears to be annotated with a 403 as well
        with pytest.raises(AuthError):
            self.create_and_send_event(room_id, self.user_d_id)


class InsuredOnlyRoomScanIgnoreInvitesTaskTestCase(FederatingModuleApiTestCase):
    """
    Test that insured only room scans are done, and required room kicks are done. Unlike
    the previous test case above, this explicitly disables considering invites as part
    of the reasons to kick all users from a room
    """

    # This test case will model being an EPA server on the federation list
    remote_pro_user = f"@mxid:{DOMAIN_IN_LIST}"
    # Our server name
    server_name_for_this_server = INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL
    # The default "fake" remote server name that has its server signing keys auto-injected
    OTHER_SERVER_NAME = DOMAIN_IN_LIST

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "epa"})
        conf["modules"][0].setdefault("config", {}).update(
            {"room_scan_run_interval": "1h"}
        )
        conf["modules"][0].setdefault("config", {}).update(
            {
                "insured_only_room_scan": {
                    "enabled": True,
                    "grace_period": "6h",
                },
                # Enabled to test broken rooms being ignored(but later purged correctly)
                # We don't worry about the grace_period as it will not apply
                "inactive_room_scan": {"enabled": True},
            }
        )

        return conf

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)

        self.user_d = self.register_user("d", "password")
        # Need the full UserID in a few places for sending a message into the room
        self.user_d_id = UserID.from_string(self.user_d)
        self.login("d", "password")

    def test_room_scan_ignores_room_with_invite_grace_period_disabled(self) -> None:
        """
        Test that a room does not have epa members kicked if the invite remains pending
        with the invite grace period disabled.
        """
        # Make a room and invite the doctor
        room_id = self.create_local_room(
            self.user_d, [self.remote_pro_user], is_public=False
        )
        assert room_id is not None, "Room should be created"

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        # Needs to be either None or False
        assert not is_room_blocked, "Room should not be blocked yet(try 1)"

        # Send a junk hex message into the room, like a sentinel
        self.create_and_send_event(room_id, self.user_d_id)

        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 2)"

        # The config gives 6 hours for leaves. Invites should count as joins and not
        # cause the users to be kicked out.
        self.reactor.advance(5 * 60 * 60)

        # Room should still exist
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked yet(try 3)"

        # Give another hour
        self.reactor.advance(60 * 60)

        # Should still be here
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked now(try 4)"

        # Give another hour
        self.reactor.advance(60 * 60)

        # Should still be here
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked now(try 5)"

        # Give another hour
        self.reactor.advance(60 * 60)

        # Should still be here
        is_room_blocked = self.get_success_or_raise(self.store.is_room_blocked(room_id))
        assert not is_room_blocked, "Room should not be blocked now(try 6)"
        # Send a junk hex message into the room. Proof the room still exists
        self.create_and_send_event(room_id, self.user_d_id)


class InactiveRoomScanTaskTestCase(FederatingModuleApiTestCase):
    """
    Test that inactive room scans are done, and subsequent room purges are run
    """

    # This test case will model being an PRO server on the federation list
    # By default we are SERVER_NAME_FROM_LIST

    # Test with one other remote PRO server and one EPA server
    # The inactive grace period is going to be 6 hours, room scans run each hour
    remote_pro_user = f"@mxid:{DOMAIN_IN_LIST}"
    remote_epa_user = f"@alice:{INSURANCE_DOMAIN_IN_LIST}"
    # The default "fake" remote server name that has its server signing keys auto-injected
    OTHER_SERVER_NAME = DOMAIN_IN_LIST

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "pro"})
        conf["modules"][0].setdefault("config", {}).update(
            {"room_scan_run_interval": "1h"}
        )
        conf["modules"][0].setdefault("config", {}).update(
            {
                "inactive_room_scan": {
                    "enabled": True,
                    "grace_period": "6h",
                }
            }
        )

        return conf

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.task_scheduler = self.hs.get_task_scheduler()

        self.user_a = self.register_user("a", "password")
        self.user_b = self.register_user("b", "password")
        self.user_c = self.register_user("c", "password")

        # Need this access token for the invite helper below
        self.access_token_a = self.login("a", "password")
        # Won't need access tokens for these users directly
        self.login("b", "password")
        self.login("c", "password")

        # OTHER_SERVER_NAME already has it's signing key injected into our database so
        # our server doesn't have to make that request. Add the other servers we will be
        # using as well
        self.inject_servers_signing_key(INSURANCE_DOMAIN_IN_LIST)

    def opinionated_join(self, room_id: str, user: str) -> None:
        """
        Helper to join a room whether this is a local or remote user
        """
        if self.module_api.is_mine(user):
            # local
            self.helper.join(room_id, user, tok=self.map_user_id_to_token[user])
        else:
            # remote
            self.send_join(user, room_id)

    def assert_task_status_for_room_is(
        self,
        room_id: str,
        task_name: str,
        status_list: list[TaskStatus],
        comment: str,
    ) -> None:
        """
        Assert that for a given room id, the Statuses listed have a single entry

        If the status_list is empty or None, there should be no tasks to find
        """
        purge_task_list = self.get_success_or_raise(
            self.task_scheduler.get_tasks(actions=[task_name], resource_id=room_id)
        )

        if status_list:
            assert (
                len(purge_task_list) > 0
            ), f"{comment} | GT status_list: {status_list}, purge_list: {purge_task_list}"
        else:
            assert (
                len(purge_task_list) == 0
            ), f"{comment} | EQ status_list: {status_list}, purge_list: {purge_task_list}"

        completed_task = [
            task for task in purge_task_list if task.status == TaskStatus.COMPLETE
        ]
        active_task = [
            task for task in purge_task_list if task.status == TaskStatus.ACTIVE
        ]
        scheduled_task = [
            task for task in purge_task_list if task.status == TaskStatus.SCHEDULED
        ]
        assert len(completed_task) == (
            1 if TaskStatus.COMPLETE in status_list else 0
        ), f"{comment} | completed {completed_task}"
        assert len(active_task) == (
            1 if TaskStatus.ACTIVE in status_list else 0
        ), f"{comment} | active {active_task}"
        assert len(scheduled_task) == (
            1 if TaskStatus.SCHEDULED in status_list else 0
        ), f"{comment} | scheduled {scheduled_task}"

    # test for private dm between two local users
    # test for private dm between a local and remote user
    # test for public room on local server
    # test for basic dm between a local and remote epa user

    # I'm not sure I like the hard coding of the user names here, but can not access
    # "self" to just reference it
    @parameterized.expand(
        [
            # (name to give the test, list of users to test with, is public room, any messages in room, should remote actually fully join?)
            (
                "private_room_2_local_users_with_messages",
                [f"@b:{SERVER_NAME_FROM_LIST}"],
                False,
                True,
                True,
            ),
            (
                "private_room_2_local_users_no_messages",
                [f"@b:{SERVER_NAME_FROM_LIST}"],
                False,
                False,
                True,
            ),
            (
                "private_room_1_local_user_1_remote_pro_user",
                [f"@mxid:{DOMAIN_IN_LIST}"],
                False,
                True,
                True,
            ),
            (
                "public_room_3_local_users",
                [f"@b:{SERVER_NAME_FROM_LIST}", f"@c:{SERVER_NAME_FROM_LIST}"],
                True,
                True,
                True,
            ),
            (
                "private_room_with_1_pro_1_epa",
                [f"@b:{SERVER_NAME_FROM_LIST}", f"@alice:{INSURANCE_DOMAIN_IN_LIST}"],
                False,
                True,
                True,
            ),
            (
                "private_room_with_1_pro_1_epa_that_does_not_join",
                [f"@b:{SERVER_NAME_FROM_LIST}", f"@alice:{INSURANCE_DOMAIN_IN_LIST}"],
                False,
                True,
                False,
            ),
        ]
    )
    def test(
        self,
        _: str,
        other_users: list[str],
        is_public: bool,
        send_messages: bool,
        remote_fully_join: bool,
    ) -> None:
        """
        Test that a room is deleted when a local PRO user and various others don't touch
        a room for "inactive_room_scan.grace_period" amount of time
        """
        # Make a room and invite the other occupant(s)
        room_id = self.create_local_room(self.user_a, [], is_public=is_public)
        assert room_id is not None, "Room should exist"

        for other_user in other_users:
            self.helper.invite(room_id, targ=other_user, tok=self.access_token_a)

        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should exist from initial get_room()"

        # other user joins
        if remote_fully_join:
            for other_user in other_users:
                self.opinionated_join(room_id, other_user)

        # Send a junk hex message into the room, this is the message the scan will find
        if send_messages:
            self.create_and_send_event(room_id, UserID.from_string(self.user_a))

        # verify there are no shutdown tasks associated with this room
        self.assert_task_status_for_room_is(
            room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [], "first check"
        )

        # wait for cleanup, should take 6 hours(based on above configuration)
        count = 0
        while True:
            count += 1
            if count == 6:
                break

            # advance() is in seconds, this should be 1 hour
            self.reactor.advance(60 * 60)

            current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
            assert (
                current_rooms is not None
            ), f"Room should still exist at count: {count}"

            self.assert_task_status_for_room_is(
                room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [], f"loop count: {count}"
            )

        # Stopped the loop above before advancing the time, so advance() for one more
        # hour, which should allow the task to be scheduled
        self.reactor.advance(60 * 60)

        # Room should still exist
        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should still exist after loop finished"

        self.assert_task_status_for_room_is(
            room_id,
            SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME,
            [TaskStatus.SCHEDULED],
            "after loop",
        )

        # The TaskScheduler has a heartbeat of 1 minute, give it that much
        self.reactor.advance(1 * 60)

        # Now the room should be gone
        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is None, f"Room should be gone now: {current_rooms}"

        # verify a scheduled task "completed" for this room
        self.assert_task_status_for_room_is(
            room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [TaskStatus.COMPLETE], "end"
        )

    def test_room_scan_purges_incomplete_inactive_rooms_after_skip(self) -> None:
        """
        Test that a partially formed room does not break the room scanner, but instead
        is skipped one time before being reconsidered. Remember, rooms scans run every
        hour by default in this TestCase
        """
        # Make the bare minimum of a room, which is basically just a row in the `rooms` table
        room_version_id = self.hs.config.server.default_room_version.identifier

        room_version = KNOWN_ROOM_VERSIONS.get(room_version_id)
        assert room_version
        room_id = self.get_success_or_raise(
            self.hs.get_room_creation_handler()._generate_and_create_room_id(
                self.user_a, is_public=False, room_version=room_version
            )
        )
        assert room_id is not None, "Room should be partially created"

        # verify there are no shutdown tasks associated with this room
        self.assert_task_status_for_room_is(
            room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [], "first check"
        )

        # Here's the first hour gone, the room scan should have ran at least once
        self.reactor.advance(60 * 60)

        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should still exist(try 1)"

        self.assert_task_status_for_room_is(
            room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [], "first check"
        )

        # One more hour and the room scan should run again, which should allow the task
        # to be scheduled for purging
        self.reactor.advance(60 * 60)

        # Room should still exist
        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should still exist(try 2)"

        self.assert_task_status_for_room_is(
            room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [], "second check"
        )

        # The TaskScheduler has a heartbeat of 1 minute, give it that much
        self.reactor.advance(60)

        # Room should still exist, since it will be ignored
        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should still exist(try 3)"

        # verify a scheduled task "completed" for this room
        self.assert_task_status_for_room_is(
            room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [], "end"
        )

    def test_scheduling_a_room_delete_is_idempotent(self) -> None:
        room_id = f"!fake_room_name:{self.server_name_for_this_server}"
        pretest_delete_tasks = self.get_success_or_raise(
            self.inv_checker.get_delete_tasks_by_room(room_id)
        )
        assert len(pretest_delete_tasks) == 0

        self.get_success_or_raise(self.inv_checker.schedule_room_for_purge(room_id))
        delete_tasks = self.get_success_or_raise(
            self.inv_checker.get_delete_tasks_by_room(room_id)
        )
        assert len(delete_tasks) == 1
        delete_task_id = delete_tasks[0].id

        self.get_success_or_raise(self.inv_checker.schedule_room_for_purge(room_id))
        second_delete_tasks = self.get_success_or_raise(
            self.inv_checker.get_delete_tasks_by_room(room_id)
        )
        assert len(second_delete_tasks) == 1
        assert delete_task_id == second_delete_tasks[0].id

    # I'm not sure I like the hard coding of the usernames here, but can not access
    # "self" to just reference it
    @parameterized.expand(
        [
            # (name to give the test, remote user to test with, is public room, should remote actually fully join?)
            (
                "private_room_1_local_user_1_remote_pro_user",
                f"@mxid:{DOMAIN_IN_LIST}",  # self.remote_pro_user
                False,
                True,
            ),
            (
                "private_room_with_1_local_pro_1_remote_epa",
                f"@alice:{INSURANCE_DOMAIN_IN_LIST}",  # self.remote_epa_user
                False,
                True,
            ),
            (
                "private_room_with_1_local_pro_1_remote_epa_that_does_not_join",
                f"@alice:{INSURANCE_DOMAIN_IN_LIST}",
                False,
                False,
            ),
        ]
    )
    def test_remote(
        self,
        _: str,
        remote_user: str,
        is_public: bool,
        finish_join: bool,
    ) -> None:
        """
        Test that a room is not deleted prematurely when a local user attempts to join a
        remote room, including when it is only an invite and no other state is
        transferred.
        """
        # Make a remote room...
        room_id = self.create_remote_room(
            remote_user,
            self.hs.config.server.default_room_version.identifier,
            is_public=is_public,
        )
        assert room_id is not None, "Room should exist"

        # ...and invite our local user...
        self.do_remote_invite(self.user_a, remote_user, room_id)

        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should exist from initial get_room()"

        # ...and maybe even make them join.
        if finish_join:
            self.do_remote_join(room_id, self.user_a)

        # Normally here we would inject a message into the room, so the scanner has
        # something to find. That infrastructure does not exist yet for the fake room.
        # We will rely on the invite and the room creation as the fallback for now.

        # verify there are no shutdown tasks associated with this room
        self.assert_task_status_for_room_is(
            room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [], "first check"
        )

        # wait for cleanup, should take 6 hours(based on above configuration)
        count = 0
        while True:
            count += 1
            if count == 6:
                break

            # advance() is in seconds, this should be 1 hour
            self.reactor.advance(60 * 60)

            current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
            assert (
                current_rooms is not None
            ), f"Room should still exist at count: {count}"

            self.assert_task_status_for_room_is(
                room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [], f"loop count: {count}"
            )

        # Stopped the loop above before advancing the time, so advance() for one more
        # hour, which should allow the task to be scheduled
        self.reactor.advance(60 * 60)

        # Room should still exist
        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is not None, "Room should still exist after loop finished"

        self.assert_task_status_for_room_is(
            room_id,
            SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME,
            [TaskStatus.SCHEDULED],
            "after loop",
        )

        # The TaskScheduler has a heartbeat of 1 minute, give it that much
        self.reactor.advance(1 * 60)

        # Now the room should be gone
        current_rooms = self.get_success_or_raise(self.store.get_room(room_id))
        assert current_rooms is None, f"Room should be gone now: {current_rooms}"

        # verify a scheduled task "completed" for this room
        self.assert_task_status_for_room_is(
            room_id, SHUTDOWN_AND_PURGE_ROOM_ACTION_NAME, [TaskStatus.COMPLETE], "end"
        )



# Copyright (C) 2020, 2024 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
from unittest.mock import AsyncMock, patch

import pytest
from pydantic import ValidationError
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from synapse_invite_checker.types import FederationDomain, FederationList
from tests.base import FederatingModuleApiTestCase
from tests.test_utils import DOMAIN_IN_LIST


class FederationDomainSchemaTest(FederatingModuleApiTestCase):
    """
    Test that the required fields for the federation list are present and parsable.
    See:
    https://github.com/gematik/api-vzd/blob/main/src/schema/FederationList.json
    for the schema to use.

    As of the time of this writing, these are fields that are required:
        domain: str
        telematikID: str
        timAnbieter: str
        isInsurance: bool

    """

    def prepare(
        self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer
    ) -> None:
        super().prepare(reactor, clock, homeserver)

    async def extract_entry_from_domainList(self, domain: str) -> FederationDomain:
        """
        Search for a specific domain in the federation list to extract it's associated
        data. Normally this additional information is not used
        """
        fed_list = await self.inv_checker._fetch_federation_list()
        assert len(fed_list.domainList) > 0

        for domain_entry in fed_list.domainList:
            if domain_entry.domain == domain:
                return domain_entry

        msg = f"Not found in federation list {domain}"
        raise AssertionError(msg)

    async def test_federation_list(self) -> None:
        """Ensure we can properly fetch the federation list"""

        fed_list = await self.inv_checker._fetch_federation_list()
        assert fed_list.allowed("timo.staging.famedly.de")

    async def test_is_insurance(self) -> None:
        """Ensure we can properly determine if a domain is insurance"""

        fed_list = await self.inv_checker._fetch_federation_list()
        assert fed_list.is_insurance("cirosec.de")

    async def test_is_domain_allowed_uses_cache(self) -> None:
        # Simple procedural test. It does not increase coverage(at the time it was
        # written), but did prove to me that it was working.
        # 1. clear the count on the mock of called instances
        # 2. get the value, which should increase counter to 1
        # 3. check that calls == 1, reset mock to 0
        # 4. get the value again, which should retrieve from cache and have a count of 0
        # self.inv_checker._raw_federation_list_fetch = AsyncMock(
        #     wraps=self.inv_checker._raw_federation_list_fetch
        # )
        with patch.object(
            self.inv_checker,
            "_raw_federation_list_fetch",
            AsyncMock(wraps=self.inv_checker._raw_federation_list_fetch),
        ) as mock_fetch:
            mock_fetch.reset_mock()
            mock_fetch.assert_not_called()

            should_be_true = await self.inv_checker.is_domain_allowed(DOMAIN_IN_LIST)
            assert should_be_true, "tested domain was not allowed but should have been"

            mock_fetch.assert_called_once()
            mock_fetch.reset_mock()

            should_still_be_true = await self.inv_checker.is_domain_allowed(
                DOMAIN_IN_LIST
            )
            assert (
                should_still_be_true
            ), "tested domain was still not allowed but should have been"
            mock_fetch.assert_not_called()

    async def test_common_fed_domain(self):
        # First test the most common FederationDomain entry
        # {
        #     "domain": "timo.staging.famedly.de",
        #     "telematikID": "1-SMC-B-Testkarte--883110000147435",
        #     "timAnbieter": "ORG-0217:BT-0158",
        #     "isInsurance": false
        # },
        test_entry = await self.extract_entry_from_domainList("timo.staging.famedly.de")
        assert test_entry.domain == "timo.staging.famedly.de"
        assert test_entry.telematikID == "1-SMC-B-Testkarte--883110000147435"
        assert test_entry.timAnbieter == "ORG-0217:BT-0158"
        assert test_entry.isInsurance is False

    async def test_insurance_fed_domain(self):
        # Then test an insurance FederationDomain entry. Want isInsurance to be True
        # {
        #     "domain": "ti-messengertest.dev.ccs.gematik.solutions",
        #     "telematikID": "5-2-KHAUS-Kornfeld01",
        #     "timAnbieter": "ORG-0001:BT-0001",
        #     "isInsurance": true
        # },

        test_entry = await self.extract_entry_from_domainList(
            "ti-messengertest.dev.ccs.gematik.solutions"
        )
        assert test_entry.domain == "ti-messengertest.dev.ccs.gematik.solutions"
        assert test_entry.telematikID == "5-2-KHAUS-Kornfeld01"
        assert test_entry.timAnbieter == "ORG-0001:BT-0001"
        assert test_entry.isInsurance is True

    async def test_illegal_fed_domain(self):
        # This test is against a FederationDomain entry with data that is counter to
        # what the schema says. In this case, 'timAnbieter' should be required but is
        # reflected as `null`
        # {
        #     "domain": "messenger.spilikin.dev",
        #     "telematikID": "1-SMC-B-Testkarte-883110000096089",
        #     "timAnbieter": null,
        #     "isInsurance": false
        # },

        test_entry = await self.extract_entry_from_domainList("messenger.spilikin.dev")
        assert test_entry.domain == "messenger.spilikin.dev"
        assert test_entry.telematikID == "1-SMC-B-Testkarte-883110000096089"
        assert test_entry.timAnbieter is None
        assert test_entry.isInsurance is False


class FederationListValidationTestCase(unittest.TestCase):
    """
    Test validating the federation list. The schema for such is declared at:
    https://github.com/gematik/api-vzd/blob/main/src/schema/FederationList.json

    """

    def test_federation_list_schema_complete(self) -> None:
        json_str = """
            {
                "domainList": [
                    {
                        "domain": "hs1",
                        "ik": [
                            "012345678"
                        ],
                        "isInsurance": false,
                        "telematikID": "fake_tid",
                        "timAnbieter": "placeholder"
                    }
                ],
                "version": 0
            }
        """
        FederationList.model_validate_json(json_str)

    def test_federation_list_schema_missing_domain(self) -> None:
        json_str = """
            {
                "domainList": [
                    {
                        "ik": [
                            "012345678"
                        ],
                        "isInsurance": false,
                        "telematikID": "fake_tid",
                        "timAnbieter": "placeholder"
                    }
                ],
                "version": 0
            }
        """
        with pytest.raises(ValidationError):
            FederationList.model_validate_json(json_str)

    def test_federation_list_schema_missing_telematik_id(self) -> None:
        json_str = """
            {
                "domainList": [
                    {
                        "domain": "hs1",
                        "ik": [
                            "012345678"
                        ],
                        "isInsurance": false,
                        "timAnbieter": "placeholder"
                    }
                ],
                "version": 0
            }
        """
        with pytest.raises(ValidationError):
            FederationList.model_validate_json(json_str)

    def test_federation_list_schema_missing_tim_anbieter(self) -> None:
        json_str = """
            {
                "domainList": [
                    {
                        "domain": "hs1",
                        "ik": [
                            "012345678"
                        ],
                        "isInsurance": false,
                        "telematikID": "fake_tid"
                    }
                ],
                "version": 0
            }
        """
        FederationList.model_validate_json(json_str)

    def test_federation_list_schema_missing_ik(self) -> None:
        json_str = """
            {
                "domainList": [
                    {
                        "domain": "hs1",
                        "isInsurance": false,
                        "telematikID": "fake_tid",
                        "timAnbieter": "placeholder"
                    }
                ],
                "version": 0
            }
        """
        FederationList.model_validate_json(json_str)

    def test_federation_list_schema_empty_ik(self) -> None:
        json_str = """
            {
                "domainList": [
                    {
                        "domain": "hs1",
                        "ik": [],
                        "isInsurance": false,
                        "telematikID": "fake_tid",
                        "timAnbieter": "placeholder"
                    }
                ],
                "version": 0
            }
        """
        FederationList.model_validate_json(json_str)

    def test_federation_list_schema_missing_is_insurance(self) -> None:
        json_str = """
            {
                "domainList": [
                    {
                        "domain": "hs1",
                        "ik": [
                            "012345678"
                        ],
                        "telematikID": "fake_tid",
                        "timAnbieter": "placeholder"
                    }
                ],
                "version": 0
            }
        """
        with pytest.raises(ValidationError):
            FederationList.model_validate_json(json_str)

    def test_federation_list_schema_minimal(self) -> None:
        """
        The Federation List schema declares that only 3 fields are required
        """
        json_str = """
            {
                "domainList": [
                    {
                        "domain": "hs1",
                        "isInsurance": false,
                        "telematikID": "fake_tid"
                    }
                ],
                "version": 0
            }
        """
        FederationList.model_validate_json(json_str)



# Copyright (C) 2025 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from http import HTTPStatus
from typing import Any
from unittest.mock import AsyncMock, Mock

from parameterized import parameterized
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor
from typing_extensions import override

from synapse_invite_checker.types import PermissionConfig, PermissionDefaultSetting
from tests.base import FederatingModuleApiTestCase
from tests.test_utils import (
    DOMAIN_IN_LIST,
    INSURANCE_DOMAIN_IN_LIST,
    INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL,
)


class IncomingRemoteJoinTestCase(FederatingModuleApiTestCase):
    """
    Test incoming remote joins(from federation) behave as expected.
    Unlike tests in test_createrooms_remote.py, these have rooms created with no invites
    """

    # By default, we are SERVER_NAME_FROM_LIST
    # server_name_for_this_server = "tim.test.gematik.de"
    # This test case will model being an PRO server on the federation list

    remote_pro_user = f"@mxid:{DOMAIN_IN_LIST}"
    remote_epa_user = f"@alice:{INSURANCE_DOMAIN_IN_LIST}"
    # The default "fake" remote server name that has its server signing keys auto-injected
    OTHER_SERVER_NAME = DOMAIN_IN_LIST

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.user_b = self.register_user("b", "password")
        self.user_c = self.register_user("c", "password")
        self.user_d = self.register_user("d", "password")
        self.access_token_a = self.login("a", "password")
        self.access_token_b = self.login("b", "password")
        self.access_token_c = self.login("c", "password")
        self.access_token_d = self.login("d", "password")

        # OTHER_SERVER_NAME already has it's signing key injected into our database so
        # our server doesn't have to make that request. Add the other servers we will be
        # using as well
        self.inject_servers_signing_key(INSURANCE_DOMAIN_IN_LIST)

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "pro"})
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": {"defaultSetting": "allow all"}}
        )
        return conf

    @parameterized.expand([("public", True), ("private", False)])
    def test_local_room_remote_epa_no_invites(self, _: str, is_public: bool) -> None:
        """
        Test _with no invites_ behavior for public and private rooms when there is an
        incoming remote user
        """
        room_id = self.create_local_room(self.user_a, [], is_public=is_public)
        assert room_id is not None, "Room should have been created"

        # public should not succeed
        # private should also not succeed
        # Since no invites occurred, we never get past make_join
        self.send_join(
            self.remote_epa_user,
            room_id,
            make_join_expected_code=HTTPStatus.FORBIDDEN,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_local_room_remote_epa_with_invites(self, _: str, is_public: bool) -> None:
        """
        Test _with invites_ behavior for public and private rooms when there is an
        incoming remote user
        """
        # Try with user "d", make a fresh room
        room_id = self.create_local_room(self.user_d, [], is_public=is_public)
        assert room_id is not None, "Room should have been created"

        # for a public room this should fail
        # for a private room this should succeed
        self.helper.invite(
            room_id,
            self.user_d,
            self.remote_epa_user,
            expect_code=HTTPStatus.FORBIDDEN if is_public else HTTPStatus.OK,
            tok=self.access_token_d,
        )

        # public room should be forbidden
        # private room should be allowed, because invite
        self.send_join(
            self.remote_epa_user,
            room_id,
            make_join_expected_code=(
                HTTPStatus.FORBIDDEN if is_public else HTTPStatus.OK
            ),
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_local_room_remote_pro_no_invites(self, _: str, is_public: bool) -> None:
        """
        Test with no invites behavior for public and private rooms when there is an
        incoming remote user
        """
        room_id = self.create_local_room(self.user_a, [], is_public=is_public)
        assert room_id is not None, "Room should have been created"

        # public should not succeed
        # private should also not succeed
        # Since no invites occurred, we never get past make_join
        self.send_join(
            self.remote_pro_user,
            room_id,
            make_join_expected_code=HTTPStatus.FORBIDDEN,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_local_room_remote_pro_with_invites(self, _: str, is_public: bool) -> None:
        """
        Test with invites behavior for public and private rooms when there is an
        incoming remote user
        """
        room_id = self.create_local_room(self.user_a, [], is_public=is_public)
        assert room_id is not None, "Room should have been created"

        # Private rooms, this should be allowed without permission
        # Public rooms, should be denied because public room
        self.helper.invite(
            room_id,
            self.user_a,
            self.remote_pro_user,
            expect_code=HTTPStatus.FORBIDDEN if is_public else HTTPStatus.OK,
            tok=self.access_token_a,
        )

        # make_join should only succeed for private rooms, and be forbidden for public
        # send_join should only succeed for private rooms
        self.send_join(
            self.remote_pro_user,
            room_id,
            make_join_expected_code=(
                HTTPStatus.FORBIDDEN if is_public else HTTPStatus.OK
            ),
        )


class DisableOverridePublicRoomFederationTestCase(FederatingModuleApiTestCase):
    """
    Test that disabling 'override_public_room_federation' allows federation to function
    Unlike tests in test_createrooms_remote.py, these have rooms created with no invites
    """

    # By default, we are SERVER_NAME_FROM_LIST
    # server_name_for_this_server = "tim.test.gematik.de"
    # This test case will model being an PRO server on the federation list

    remote_pro_user = f"@mxid:{DOMAIN_IN_LIST}"
    remote_epa_user = f"@alice:{INSURANCE_DOMAIN_IN_LIST}"
    # The default "fake" remote server name that has its server signing keys auto-injected
    OTHER_SERVER_NAME = DOMAIN_IN_LIST

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.access_token_a = self.login("a", "password")
        self.user_d = self.register_user("d", "password")
        self.access_token_d = self.login("d", "password")

        # OTHER_SERVER_NAME already has it's signing key injected into our database so
        # our server doesn't have to make that request. Add the other servers we will be
        # using as well
        self.inject_servers_signing_key(INSURANCE_DOMAIN_IN_LIST)

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()

        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update(
            {"override_public_room_federation": False}
        )

        return conf

    def test_room_federation(self) -> None:
        """
        Test that the local server can successfully allow joining a remote room when
        there are no invites
        """
        room_id = self.create_local_room(self.user_a, [], is_public=True)
        assert room_id is not None, "Room should have been created"

        # make_join should succeed, as the override was blocked
        self.send_join(
            self.remote_pro_user,
            room_id,
        )


class OutgoingEPARemoteJoinTestCase(FederatingModuleApiTestCase):
    """
    Test for behavior when joining a remote room
    """

    server_name_for_this_server = INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL
    # This test case will model being an EPA server on the federation list
    # By default we are SERVER_NAME_FROM_LIST

    # Test with one other remote PRO server and one EPA server
    remote_pro_user = f"@mxid:{DOMAIN_IN_LIST}"
    remote_epa_user = f"@alice:{INSURANCE_DOMAIN_IN_LIST}"
    # The default "fake" remote server name that has its server signing keys auto-injected
    OTHER_SERVER_NAME = DOMAIN_IN_LIST

    @override
    def make_homeserver(self, reactor: MemoryReactor, clock: Clock) -> HomeServer:
        # Mock out the calls over federation.
        self.fed_transport_client = Mock(spec=["send_transaction"])
        self.fed_transport_client.send_transaction = AsyncMock(return_value={})

        return self.setup_test_homeserver(
            # Masquerade as a domain found on the federation list, then we can pass
            # tests that verify that fact
            self.server_name_for_this_server,
            federation_transport_client=self.fed_transport_client,
        )

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.user_d = self.register_user("d", "password")
        self.login("a", "password")
        self.login("d", "password")

        # OTHER_SERVER_NAME already has it's signing key injected into our database so
        # our server doesn't have to make that request. Add the other servers we will be
        # using as well
        self.inject_servers_signing_key(INSURANCE_DOMAIN_IN_LIST)

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "epa"})
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": {"defaultSetting": "allow all"}}
        )
        return conf

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_pro_no_invites(self, _: str, is_public: bool) -> None:
        """
        Test that the local server can successfully block joining a remote room when
        there are no invites
        """
        remote_room_id = self.create_remote_room(self.remote_pro_user, "10", is_public)
        assert remote_room_id is not None

        # Public rooms should fail.
        # Private rooms should also fail because no invite. Should be a 403
        self.do_remote_join(
            remote_room_id, self.user_a, expected_code=HTTPStatus.FORBIDDEN
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_pro_with_invites(self, _: str, is_public: bool) -> None:
        """
        Test that the local server can successfully join a remote PRO server room, when
        appropriate, if there are invites
        """
        remote_room_id = self.create_remote_room(self.remote_pro_user, "10", is_public)
        assert remote_room_id is not None

        # This should be enough to inject the "fact" we got an invite, and should
        # allow private room joining
        self.do_remote_invite(self.user_a, self.remote_pro_user, remote_room_id)

        # Public rooms should fail with a 403. Private rooms should succeed, because of
        # the invite
        self.do_remote_join(
            remote_room_id,
            self.user_a,
            expected_code=HTTPStatus.FORBIDDEN if is_public else None,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_pro_with_epa_invite_second_epa_user_fails_join_with_no_invite(
        self, _: str, is_public: bool
    ) -> None:
        """
        Test that the local EPA server can successfully join a remote PRO server room,
        when appropriate. The first local user gets an invite. The second doesn't and
        the join fails correctly
        """
        remote_room_id = self.create_remote_room(self.remote_pro_user, "10", is_public)
        assert remote_room_id is not None

        # This should be enough to inject the "fact" we got an invite, and should
        # allow private room joining
        self.do_remote_invite(self.user_a, self.remote_pro_user, remote_room_id)

        # Public rooms should fail with a 403. Private rooms should succeed, because of
        # the invite
        self.do_remote_join(
            remote_room_id,
            self.user_a,
            expected_code=HTTPStatus.FORBIDDEN if is_public else None,
        )
        # Now try and join the room with our second local user, should fail because no invite
        self.do_remote_join(
            remote_room_id,
            self.user_d,
            expected_code=HTTPStatus.FORBIDDEN,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_epa_no_invites(self, _: str, is_public: bool) -> None:
        """
        Test that the local server can successfully block a remote EPA server room when
        there are no invites
        """
        remote_room_id = self.create_remote_room(self.remote_epa_user, "10", is_public)
        assert remote_room_id is not None

        # In both cases, this raises. Neither private nor public rooms are allowed.
        # Public, because they are denied without invites, and private because the
        # there was no invite
        self.do_remote_join(
            remote_room_id,
            self.user_a,
            expected_code=HTTPStatus.FORBIDDEN,
        )

        # Try with a different local user, one with no visibility
        # Public rooms should fail. Private rooms should also fail because no invite
        self.do_remote_join(
            remote_room_id,
            self.user_d,
            expected_code=HTTPStatus.FORBIDDEN,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_epa_with_invites_blocked(
        self, _: str, is_public: bool
    ) -> None:
        """
        Test that the local server can successfully block a remote EPA server room, when
        an invite is blocked from lack of permissions
        """
        remote_room_id = self.create_remote_room(self.remote_epa_user, "10", is_public)
        assert remote_room_id is not None

        self.set_permissions_for_user(
            self.user_a,
            PermissionConfig(defaultSetting=PermissionDefaultSetting.BLOCK_ALL),
        )

        # This should be enough to inject the "fact" we got an invite, and should
        # allow private room joining. However, user "a" has 'block all' permissions so
        # the invite should always fail
        self.do_remote_invite(
            self.user_a,
            self.remote_epa_user,
            remote_room_id,
            expect_code=HTTPStatus.FORBIDDEN,
        )

        # In both cases, this raises. Neither private nor public rooms are allowed.
        # Public, because they are denied without invites, and private because the
        # invite failed above
        self.do_remote_join(
            remote_room_id,
            self.user_a,
            expected_code=HTTPStatus.FORBIDDEN,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_epa_with_invites_allowed(
        self, _: str, is_public: bool
    ) -> None:
        """
        Test that the local server can successfully join a remote EPA server room, when
        an invite is allowed
        """
        remote_room_id = self.create_remote_room(self.remote_epa_user, "10", is_public)
        assert remote_room_id is not None

        # User 'd' must grant their permission
        self.set_permissions_for_user(
            self.user_d,
            PermissionConfig(
                defaultSetting=PermissionDefaultSetting.BLOCK_ALL,
                userExceptions={self.remote_epa_user: {}},
            ),
        )

        # This should be enough to inject the "fact" we got an invite, and should
        # fail both public and private room joining because two EPA servers
        self.do_remote_invite(
            self.user_d,
            self.remote_epa_user,
            remote_room_id,
            expect_code=HTTPStatus.FORBIDDEN,
        )

        # Both public and private should fail, because two EPA servers(and no invites)
        self.do_remote_join(
            remote_room_id,
            self.user_d,
            expected_code=HTTPStatus.FORBIDDEN,
        )


class OutgoingPRORemoteJoinTestCase(FederatingModuleApiTestCase):
    """
    Test for behavior when joining a remote room
    """

    # server_name_for_this_server = "tim.test.gematik.de"
    # This test case will model being an PRO server on the federation list
    # By default we are SERVER_NAME_FROM_LIST

    # Test with one other remote PRO server and one EPA server
    remote_pro_user = f"@mxid:{DOMAIN_IN_LIST}"
    remote_epa_user = f"@alice:{INSURANCE_DOMAIN_IN_LIST}"
    # The default "fake" remote server name that has its server signing keys auto-injected
    OTHER_SERVER_NAME = DOMAIN_IN_LIST

    @override
    def make_homeserver(self, reactor: MemoryReactor, clock: Clock) -> HomeServer:
        # Mock out the calls over federation.
        self.fed_transport_client = Mock(spec=["send_transaction"])
        self.fed_transport_client.send_transaction = AsyncMock(return_value={})

        return self.setup_test_homeserver(
            # Masquerade as a domain found on the federation list, then we can pass
            # tests that verify that fact
            self.server_name_for_this_server,
            federation_transport_client=self.fed_transport_client,
        )

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.user_d = self.register_user("d", "password")
        self.login("a", "password")
        self.login("d", "password")

        # OTHER_SERVER_NAME already has it's signing key injected into our database so
        # our server doesn't have to make that request. Add the other servers we will be
        # using as well
        self.inject_servers_signing_key(INSURANCE_DOMAIN_IN_LIST)

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "pro"})
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": {"defaultSetting": "allow all"}}
        )
        return conf

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_pro_no_invites(self, _: str, is_public: bool) -> None:
        """
        Test that the local server can successfully block joining a remote room when
        there are no invites
        """
        remote_room_id = self.create_remote_room(self.remote_pro_user, "10", is_public)
        assert remote_room_id is not None

        # Public rooms should fail.
        # Private rooms should also fail because no invite. Should be a 403
        self.do_remote_join(
            remote_room_id, self.user_a, expected_code=HTTPStatus.FORBIDDEN
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_pro_with_invites(self, _: str, is_public: bool) -> None:
        """
        Test that the local server can successfully join a remote PRO server room, when
        appropriate, if there are invites
        """
        remote_room_id = self.create_remote_room(self.remote_pro_user, "10", is_public)
        assert remote_room_id is not None

        # This should be enough to inject the "fact" we got an invite, and should
        # allow private room joining
        self.do_remote_invite(self.user_a, self.remote_pro_user, remote_room_id)

        # Public rooms should fail with a 403. Private rooms should succeed, because of
        # the invite
        self.do_remote_join(
            remote_room_id,
            self.user_a,
            expected_code=HTTPStatus.FORBIDDEN if is_public else None,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_pro_with_pro_invite_second_pro_user_fails_join_with_no_invite(
        self, _: str, is_public: bool
    ) -> None:
        """
        Test that the local PRO server can successfully join a remote PRO server room,
        when appropriate. The first user gets an invite, the second user does not and
        the join fails correctly
        """
        remote_room_id = self.create_remote_room(self.remote_pro_user, "10", is_public)
        assert remote_room_id is not None

        # This should be enough to inject the "fact" we got an invite, and should
        # allow private room joining
        self.do_remote_invite(self.user_a, self.remote_pro_user, remote_room_id)

        # Public rooms should fail with a 403. Private rooms should succeed, because of
        # the invite
        self.do_remote_join(
            remote_room_id,
            self.user_a,
            expected_code=HTTPStatus.FORBIDDEN if is_public else None,
        )
        # Now try and join the room with our second local user, should fail:
        # public room: Because there is no state to access(the first user did not succeed)
        # private room: Because there was no invite
        self.do_remote_join(
            remote_room_id,
            self.user_d,
            expected_code=HTTPStatus.FORBIDDEN,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_epa_no_invites(self, _: str, is_public: bool) -> None:
        """
        Test that the local server can successfully block a remote EPA server room when
        there are no invites
        """
        remote_room_id = self.create_remote_room(self.remote_epa_user, "10", is_public)
        assert remote_room_id is not None

        # In both cases, this raises. Neither private nor public rooms are allowed.
        # Public, because they are denied without invites, and private because the
        # there was no invite
        self.do_remote_join(
            remote_room_id,
            self.user_a,
            expected_code=HTTPStatus.FORBIDDEN,
        )

        # Try with a different local user, one with no visibility
        # Public rooms should fail. Private rooms should also fail because no invite
        self.do_remote_join(
            remote_room_id,
            self.user_d,
            expected_code=HTTPStatus.FORBIDDEN,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_epa_with_invites_blocked(
        self, _: str, is_public: bool
    ) -> None:
        """
        Test that the local server can successfully block a remote EPA server room, when
        an invite is blocked from lack of permissions
        """
        remote_room_id = self.create_remote_room(self.remote_epa_user, "10", is_public)
        assert remote_room_id is not None

        self.set_permissions_for_user(
            self.user_a,
            PermissionConfig(defaultSetting=PermissionDefaultSetting.BLOCK_ALL),
        )

        # This should be enough to inject the "fact" we got an invite, and should
        # allow private room joining. However, user "a" has 'block all' permissions so
        # the invite should always fail
        self.do_remote_invite(
            self.user_a,
            self.remote_epa_user,
            remote_room_id,
            expect_code=HTTPStatus.FORBIDDEN,
        )

        # In both cases, this raises. Neither private nor public rooms are allowed.
        # Public, because they are denied without invites, and private because the
        # invite failed above
        self.do_remote_join(
            remote_room_id,
            self.user_a,
            expected_code=HTTPStatus.FORBIDDEN,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_remote_room_epa_with_invites_allowed(
        self, _: str, is_public: bool
    ) -> None:
        """
        Test that the local server can successfully join a remote EPA server room, when
        an invite is allowed
        """
        remote_room_id = self.create_remote_room(self.remote_epa_user, "10", is_public)
        assert remote_room_id is not None

        # User 'd' must grant their permission
        self.set_permissions_for_user(
            self.user_d,
            PermissionConfig(
                defaultSetting=PermissionDefaultSetting.BLOCK_ALL,
                userExceptions={self.remote_epa_user: {}},
            ),
        )

        # This should be enough to inject the "fact" we got an invite, and should
        # allow private room joining.
        self.do_remote_invite(
            self.user_d,
            self.remote_epa_user,
            remote_room_id,
            expect_code=HTTPStatus.OK,
        )

        # Public rooms should fail. Private rooms should succeed, but only because of
        # the invite
        self.do_remote_join(
            remote_room_id,
            self.user_d,
            expected_code=HTTPStatus.FORBIDDEN if is_public else None,
        )
